import copy
import pprint
import numpy as np
import warnings
from yggdrasil import backwards, tools, units, serialize
from yggdrasil.metaschema import get_metaschema
from yggdrasil.metaschema.datatypes import (
    guess_type_from_obj, get_type_from_def, get_type_class, compare_schema)
from yggdrasil.metaschema.properties.ScalarMetaschemaProperties import (
    definition2dtype, _flexible_types)
from yggdrasil.metaschema.datatypes.MetaschemaType import MetaschemaType
from yggdrasil.metaschema.datatypes.ArrayMetaschemaType import (
    OneDArrayMetaschemaType)


class SerializeBase(tools.YggClass):
    r"""Base class for serializing/deserializing a Python object into/from a
    bytes message.

    Args:
        newline (str, optional): One or more characters indicating a newline.
            Defaults to '\n'.
        comment (str, optional): One or more characters indicating a comment.
            Defaults to '# '.
        datatype (schema, optional): JSON schema defining the type of object
            that the serializer will be used to serialize/deserialize. Defaults
            to default_datatype.
        **kwargs: Additional keyword args are processed as part of the type
            definition.

    Attributes:
        initialized (bool): True if the serializer has been initialized either
            by input arguments specifying the type or by infering the type from
            a processed message.

    Class Attributes:
        has_header (bool): True if the serialization has a header when written
            to a file.
        default_read_meth (str): Default method that data should be read from
            a file for deserialization.
        is_framed (bool): True if the serialization has a frame allowing
            multiple serialized objects to be recovered from a single message.
        concats_as_str (bool): True if serialized objects can be concatenated
            directly as strings.
        encoded_datatype (schema): JSON schema defining the type of object
            produced by the class's func_serialize method. For most classes
            this will be {'type': 'bytes'}, indicating that the method will
            produce bytes suitable for serialization.

    """

    _seritype = None
    _schema_type = 'serializer'
    _schema_subtype_key = 'seritype'
    _schema_requried = ['seritype']
    _schema_properties = {
        'seritype': {'type': 'string',
                     'default': 'default',
                     'description': ('Serializer type.')},
        'newline': {'type': 'string',
                    'default': backwards.as_str(serialize._default_newline)},
        'comment': {'type': 'string',
                    'default': backwards.as_str(serialize._default_comment)},
        'datatype': {'type': 'schema'}}
    _oldstyle_kws = ['format_str', 'field_names', 'field_units', 'as_array']
    _attr_conv = ['newline', 'comment']
    default_datatype = {'type': 'bytes'}
    default_encoded_datatype = {'type': 'bytes'}
    has_header = False
    default_read_meth = 'read'
    is_framed = False
    concats_as_str = True
    
    def __init__(self, **kwargs):
        if isinstance(kwargs.get('datatype', None), MetaschemaType):
            self.datatype = kwargs.pop('datatype')
        super(SerializeBase, self).__init__(**kwargs)
        kwargs = self.extra_kwargs
        self.extra_kwargs = {}
        # Set defaults
        if self.datatype is None:
            self.datatype = self.default_datatype
        elif ((isinstance(self.datatype, dict)
               and (self.datatype != self.default_datatype))):
            kwargs['datatype'] = self.datatype
        # Update typedef
        self.initialized = False
        if isinstance(self.datatype, dict):
            self.datatype = get_type_from_def(self.default_datatype,
                                              dont_complete=True)
        if getattr(self, 'encoded_datatype', None) is None:
            self.encoded_datatype = self.default_encoded_datatype
        if isinstance(self.encoded_datatype, dict):
            self.encoded_datatype = get_type_from_def(self.encoded_datatype,
                                                      dont_complete=True)
        self.update_serializer(**kwargs)
        self.initialized = self.is_initialized()
        
    def is_initialized(self):
        r"""Determine if the serializer has been initialized by comparing the
        current datatype against the default for the class.

        Returns:
            bool: True if the current datatype is different than the default,
               False otherwise.

        """
        return (self.typedef != self.default_datatype)

    @staticmethod
    def before_registration(cls):
        r"""Operations that should be performed to modify class attributes prior
        to registration."""
        tools.YggClass.before_registration(cls)
        # If the serialization cannot be concatenated, then it is not framed by
        # definition and would be meaningless if read-in incrementally
        if not cls.concats_as_str:
            assert(not cls.is_framed)
            assert(cls.default_read_meth == 'read')
        
    @classmethod
    def object2dict(cls, obj, **kwargs):
        r"""Convert a message object into a dictionary.

        Args:
            obj (object): Object that would be serialized by this class and
                should be returned in a dictionary form.
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            dict: Dictionary version of the provided object.

        """
        return {'f0': obj}

    @classmethod
    def object2array(cls, obj, **kwargs):
        r"""Convert a message object into an array.

        Args:
            obj (object): Object that would be serialized by this class and
                should be returned in an array form.
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            np.array: Array version of the provided object or None if one cannot
               be created.

        """
        return None

    @classmethod
    def concatenate(cls, objects, **kwargs):
        r"""Concatenate objects to get object that would be recieved if
        the concatenated serialization were deserialized.

        Args:
            objects (list): Objects to be concatenated.
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            list: Set of objects that results from concatenating those provided.

        """
        return objects

    @classmethod
    def get_testing_options(cls, table_example=False, array_columns=False,
                            include_oldkws=False):
        r"""Method to return a dictionary of testing options for this class.

        Arguments:
            table_example (bool, optional): If True, the returned options will
                be for an array of elements representing a table-like structure.
                Defaults to False.
            array_columns (bool, optional): If True, table_example is set to
                True and the returned options will be for an array data type
                where each element is an array representing a column Defaults to
                False.
            include_oldkws (bool, optional): If True, old-style keywords will be
                added to the returned options. This will only have an effect if
                table_example is True. Defaults to False.

        Returns:
            dict: Dictionary of variables to use for testing. Key/value pairs:

            * kwargs (dict): Keyword arguments for comms tested with the
              provided content.
            * empty (object): Object produced from deserializing an empty
              message.
            * objects (list): List of objects to be serialized/deserialized.
              extra_kwargs (dict): Extra keyword arguments not used to
              construct type definition.
            * typedef (dict): Type definition resulting from the supplied
              kwargs.
            * dtype (np.dtype): Numpy data types that is consistent with the
              determined type definition.
            * contents (bytes): Concatenated serialization that will result from
              deserializing the serialized objects.
            * contents_recv (list): List of objects that would be deserialized
              from contents.

        """
        if array_columns:
            table_example = True
        if table_example:
            rows = [(b'one', np.int32(1), 1.0),
                    (b'two', np.int32(2), 2.0),
                    (b'three', np.int32(3), 3.0)]
            out = {'kwargs': {}, 'empty': [], 'dtype': None,
                   'extra_kwargs': {},
                   'typedef': {'type': 'array',
                               'items': [{'type': 'bytes',
                                          'units': 'n/a', 'title': 'name'},
                                         {'type': 'int', 'precision': 32,
                                          'units': 'umol', 'title': 'count'},
                                         {'type': 'float', 'precision': 64,
                                          'units': 'cm', 'title': 'size'}]},
                   'contents': (b'# name\tcount\tsize\n'
                                + b'# n/a\tumol\tcm\n'
                                + b'# %5s\t%d\t%f\n'
                                + b'  one\t1\t1.000000\n'
                                + b'  two\t2\t2.000000\n'
                                + b'three\t3\t3.000000\n'
                                + b'  one\t1\t1.000000\n'
                                + b'  two\t2\t2.000000\n'
                                + b'three\t3\t3.000000\n'),
                   'objects': 2 * rows,
                   'field_names': ['name', 'count', 'size'],
                   'field_units': ['n/a', 'umol', 'cm']}
            if include_oldkws:
                out['kwargs'].update({'format_str': '%5s\t%d\t%f\n',
                                      'field_names': ['name', 'count', 'size'],
                                      'field_units': ['n/a', 'umol', 'cm']})
                out['extra_kwargs'].update(
                    {'format_str': backwards.as_str(out['kwargs']['format_str'])})
            if array_columns:
                out['kwargs']['as_array'] = True
                dtype = np.dtype(
                    {'names': out['field_names'],
                     'formats': ['%s5' % backwards.np_dtype_str, 'i4', 'f8']})
                out['dtype'] = dtype
                arr = np.array(rows, dtype=dtype)
                lst = [units.add_units(arr[n], u) for n, u
                       in zip(out['field_names'], out['field_units'])]
                out['objects'] = [lst, lst]
                for x in out['typedef']['items']:
                    x['subtype'] = x['type']
                    x['type'] = '1darray'
                    if x['title'] == 'name':
                        x['precision'] = 40
        else:
            out = {'kwargs': {}, 'empty': b'', 'dtype': None,
                   'typedef': cls.default_datatype,
                   'extra_kwargs': {},
                   'objects': [b'Test message\n', b'Test message 2\n']}
            out['contents'] = b''.join(out['objects'])
        return out
        
    @property
    def read_meth(self):
        r"""str: Method that should be used to read data for deserialization."""
        return self.default_read_meth
        
    @classmethod
    def seri_kws(cls):
        r"""Get a list of valid keyword arguments."""
        return list(set(list(cls._schema_properties.keys()) + cls._oldstyle_kws))

    @property
    def typedef(self):
        r"""dict: Type definition."""
        return copy.deepcopy(self.datatype._typedef)

    @property
    def encoded_typedef(self):
        r"""dict: Type definition for encoded data objects."""
        return self.encoded_datatype._typedef

    @property
    def input_kwargs(self):
        r"""dict: Get the input keyword arguments used to create this class."""
        out = {}
        for k in self._schema_properties.keys():
            v = getattr(self, k, None)
            if v is not None:
                out[k] = copy.deepcopy(v)
        for k in self._attr_conv:
            if (k in out) and isinstance(out[k], backwards.string_types):
                out[k] = backwards.as_str(out[k])
        return out
        
    @property
    def serializer_info(self):
        r"""dict: Serializer info."""
        out = copy.deepcopy(self.extra_kwargs)
        for k in self._schema_properties.keys():
            if k in ['datatype']:
                continue
            v = getattr(self, k, None)
            if v is not None:
                out[k] = copy.deepcopy(v)
        for k in out.keys():
            v = out[k]
            if isinstance(v, backwards.string_types):
                out[k] = backwards.as_str(v)
            elif isinstance(v, (list, tuple)):
                out[k] = []
                for x in v:
                    out[k].append(backwards.as_str(x, allow_pass=True))
            else:
                out[k] = v
        return out

    @property
    def empty_msg(self):
        r"""obj: Object indicating empty message."""
        return self.datatype._empty_msg

    # def is_empty(self, obj):
    #     r"""Determine if an object represents an empty message for this serializer.

    #     Args:
    #         obj (object): Object to test.

    #     Returns:
    #         bool: True if the object is empty, False otherwise.

    #     """
    #     emsg = self.empty_msg
    #     return (isinstance(obj, type(emsg)) and (obj == emsg))

    def get_field_names(self, as_bytes=False):
        r"""Get the field names for an array of fields.

        Args:
            as_bytes (bool, optional): If True, the field names will be returned
                as bytes. If False the field names will be returned as unicode.
                Defaults to False.

        Returns:
            list: Names for each field in the data type.

        """
        if getattr(self, 'field_names', None) is not None:
            out = self.field_names
        elif self.typedef['type'] != 'array':
            out = None
        elif isinstance(self.typedef['items'], dict):  # pragma: debug
            raise Exception("Variable number of items not yet supported.")
        elif isinstance(self.typedef['items'], list):
            out = []
            any_names = False
            for i, x in enumerate(self.typedef['items']):
                out.append(x.get('title', 'f%d' % i))
                if len(x.get('title', '')) > 0:
                    any_names = True
            # Don't use field names if they are all defaults
            if not any_names:
                out = None
        if (out is not None):
            if as_bytes:
                out = [backwards.as_bytes(x) for x in out]
            else:
                out = [backwards.as_str(x) for x in out]
        return out

    def get_field_units(self, as_bytes=False):
        r"""Get the field units for an array of fields.

        Args:
            as_bytes (bool, optional): If True, the field units will be returned
                as bytes. If False the field units will be returned as unicode.
                Defaults to False.

        Returns:
            list: Units for each field in the data type.

        """
        if self.typedef['type'] != 'array':
            return None
        if getattr(self, 'field_units', None) is not None:
            out = self.field_units
        elif isinstance(self.typedef['items'], dict):  # pragma: debug
            raise Exception("Variable number of items not yet supported.")
        elif isinstance(self.typedef['items'], list):
            out = []
            any_units = False
            for i, x in enumerate(self.typedef['items']):
                out.append(x.get('units', ''))
                if len(x.get('units', '')) > 0:
                    any_units = True
            # Don't use field units if they are all defaults
            if not any_units:
                out = None
        if (out is not None):
            if as_bytes:
                out = [backwards.as_bytes(x) for x in out]
            else:
                out = [backwards.as_str(x) for x in out]
        return out

    @property
    def numpy_dtype(self):
        r"""np.dtype: Corresponding structured data type. Will be None unless the
        type is an array of 1darrays."""
        out = None
        if (self.typedef['type'] == 'array') and ('items' in self.typedef):
            if isinstance(self.typedef['items'], dict):
                as_array = (self.typedef['items']['type'] in ['1darray', 'ndarray'])
                if as_array:
                    out = definition2dtype(self.typedef['items'])
            elif isinstance(self.typedef['items'], (list, tuple)):
                as_array = True
                dtype_list = []
                field_names = []
                for i, x in enumerate(self.typedef['items']):
                    if x['type'] != '1darray':
                        as_array = False
                        break
                    dtype_list.append(definition2dtype(x))
                    field_names.append(x.get('title', 'f%d' % i))
                if as_array:
                    out = np.dtype(dict(names=field_names, formats=dtype_list))
        return out

    def initialize_from_message(self, msg, **metadata):
        r"""Initialize the serializer based on recieved message.

        Args:
            msg (object): Message that serializer should be initialized from.
            **kwargs: Additional keyword arguments are treated as metadata that
                may contain additional information for initializing the serializer.

        """
        if ((self.initialized or metadata.get('raw', False)
             or metadata.get('incomplete', False))):
            return
        cls = guess_type_from_obj(msg)
        typedef = cls.encode_type(msg)
        typedef = cls.extract_typedef(typedef)
        metadata.update(typedef)
        self.initialize_serializer(metadata)

    def initialize_serializer(self, metadata, extract=False):
        r"""Initialize a serializer based on received metadata. This method will
        exit early if the serializer has already been intialized.

        Args:
            metadata (dict): Header information including type info that should be
                used to initialize the serializer class.
            extract (bool, optional): If True, the type will be defined using a
                subset of the type information in metadata. If False, all of the
                type information will be used. Defaults to False.

        """
        if ((self.initialized or metadata.get('raw', False)
             or metadata.get('incomplete', False))):
            return
        self.update_serializer(extract=extract, **metadata)
        self.initialized = (self.typedef != self.default_datatype)

    def update_serializer(self, extract=False, skip_type=False, **kwargs):
        r"""Update serializer with provided information.

        Args:
            extract (bool, optional): If True, the updated typedef will be
                the bare minimum as extracted from total set of provided
                keywords, otherwise the entire set will be sued. Defaults to
                False.
            skip_type (bool, optional): If True, everything is updated except
                the data type. Defaults to False.
            **kwargs: Additional keyword arguments are processed as part of
                they type definition and are parsed for old-style keywords.

        Raises:
            RuntimeError: If there are keywords that are not valid typedef
                keywords (currect or old-style).

        """
        old_datatype = None
        if self.initialized:
            old_datatype = copy.deepcopy(self.datatype)
        _metaschema = get_metaschema()
        # Raise an error if the types are not compatible
        seritype = kwargs.pop('seritype', self.seritype)
        if (seritype != self._seritype) and (seritype != 'default'):  # pragma: debug
            raise Exception("Cannot change types form %s to %s." %
                            (self._seritype, seritype))
        # Remove metadata keywords unrelated to serialization
        # TODO: Find a better way of tracking these
        _remove_kws = ['body', 'address', 'size', 'id', 'incomplete', 'raw',
                       'commtype', 'filetype', 'response_address', 'request_id',
                       'append', 'in_temp', 'is_series', 'working_dir', 'fmts',
                       'model_driver', 'env', 'send_converter', 'recv_converter',
                       'typedef_base']
        kws = list(kwargs.keys())
        for k in kws:
            if (k in _remove_kws) or k.startswith('zmq'):
                kwargs.pop(k)
        # Set attributes and remove unused metadata keys
        for k in self._schema_properties.keys():
            if (k in kwargs) and (k != 'datatype'):
                setattr(self, k, kwargs.pop(k))
        # Create preliminary typedef
        typedef = kwargs.pop('datatype', {})
        for k in _metaschema['properties'].keys():
            if k in kwargs:
                typedef[k] = kwargs.pop(k)
        # Update extra keywords
        if (len(kwargs) > 0):
            self.extra_kwargs.update(kwargs)
            self.debug("Extra kwargs: %s" % str(self.extra_kwargs))
        # Update type
        if not skip_type:
            # Update typedef from oldstyle keywords in extra_kwargs
            typedef = self.update_typedef_from_oldstyle(typedef)
            if typedef.get('type', None):
                if extract:
                    cls = get_type_class(typedef['type'])
                    typedef = cls.extract_typedef(typedef)
                self.datatype = get_type_from_def(typedef)
            # Check to see if new datatype is compatible with new one
            if old_datatype is not None:
                errors = list(compare_schema(self.typedef, old_datatype._typedef) or ())
                if errors:
                    raise RuntimeError(
                        ("Updated datatype is not compatible with the existing one."
                         + "    New:\n%s\nOld:\n%s\n") % (
                             pprint.pformat(self.typedef),
                             pprint.pformat(old_datatype._typedef)))
        # Enfore that strings used with messages are in bytes
        for k in self._attr_conv:
            v = getattr(self, k, None)
            if isinstance(v, backwards.string_types):
                setattr(self, k, backwards.as_bytes(v))

    def update_typedef_from_oldstyle(self, typedef):
        r"""Update a given typedef using an old, table-style serialization spec.
        Existing typedef values are not overwritten and warnings are raised if the
        provided serialization spec is not compatible with the type definition.

        Args:
            typedef (dict): Type definition to update.

        Returns:
            dict: Updated typedef.

        """
        for k in self._oldstyle_kws:
            used = []
            updated = []
            v = self.extra_kwargs.get(k, getattr(self, k, None))
            if v is None:
                continue
            # Check status
            if ((k != 'format_str') and (typedef.get('type', None) != 'array')):
                continue
            # Key specific changes to type
            if k == 'format_str':
                v = backwards.as_str(v)
                fmts = serialize.extract_formats(v)
                if 'type' in typedef:
                    if (typedef.get('type', None) == 'array'):
                        assert(len(typedef.get('items', [])) == len(fmts))
                        # if len(typedef.get('items', [])) != len(fmts):
                        #     warnings.warn(("Number of items in typedef (%d) doesn't"
                        #                    + "match the number of formats (%d).")
                        #                   % (len(typedef.get('items', [])), len(fmts)))
                    continue
                as_array = self.extra_kwargs.get('as_array',
                                                 getattr(self, 'as_array', False))
                typedef.update(type='array', items=[])
                for i, fmt in enumerate(fmts):
                    nptype = serialize.cformat2nptype(fmt)
                    itype = OneDArrayMetaschemaType.encode_type(np.ones(1, nptype))
                    itype = OneDArrayMetaschemaType.extract_typedef(itype)
                    if (fmt == '%s') and ('precision' in itype):
                        del itype['precision']
                    if as_array:
                        itype['type'] = '1darray'
                    else:
                        itype['type'] = itype.pop('subtype')
                        if (((itype['type'] in _flexible_types)
                             and ('precision' in itype))):
                            del itype['precision']
                    typedef['items'].append(itype)
                used.append('as_array')
                updated.append('format_str')
            elif k == 'as_array':
                # Can only be used in conjunction with format_str
                pass
            elif k in ['field_names', 'field_units']:
                v = [backwards.as_str(x) for x in v]
                if k == 'field_names':
                    tk = 'title'
                else:
                    tk = 'units'
                if isinstance(typedef['items'], dict):
                    typedef['items'] = [copy.deepcopy(typedef['items'])
                                        for _ in range(len(v))]
                assert(len(v) == len(typedef.get('items', [])))
                # if len(v) != len(typedef.get('items', [])):
                #     warnings.warn('%d %ss provided, but only %d items in typedef.'
                #                   % (len(v), k, len(typedef.get('items', []))))
                #     continue
                all_updated = True
                for iv, itype in zip(v, typedef.get('items', [])):
                    if tk in itype:
                        all_updated = False
                    itype.setdefault(tk, iv)
                if all_updated:
                    used.append(k)
                updated.append(k)  # Won't change anything unless its an attribute
            else:  # pragma: debug
                raise ValueError(
                    "Unrecognized table-style specification keyword: '%s'." % k)
            for rk in used:
                if rk in self.extra_kwargs:
                    del self.extra_kwargs[rk]
            for rk in updated:
                if rk in self.extra_kwargs:
                    self.extra_kwargs[rk] = v
                elif hasattr(self, rk):
                    setattr(self, rk, v)
        return typedef

    def func_serialize(self, args):
        r"""Serialize a message.

        Args:
            args: List of arguments to be formatted or numpy array to be
                serialized.

        Returns:
            bytes, str: Serialized message.

        """
        raise NotImplementedError("func_serialize not implemented.")

    def func_deserialize(self, msg):
        r"""Deserialize a message.

        Args:
            msg: Message to be deserialized.

        Returns:
            obj: Deserialized message.

        """
        raise NotImplementedError("func_deserialize not implemented.")
    
    def serialize(self, args, header_kwargs=None, add_serializer_info=False,
                  no_metadata=False):
        r"""Serialize a message.

        Args:
            args (obj): List of arguments to be formatted or a ready made message.
            header_kwargs (dict, optional): Keyword arguments that should be
                added to the header. Defaults to None and no header is added.
            add_serializer_info (bool, optional): If True, serializer information
                will be added to the metadata. Defaults to False.
            no_metadata (bool, optional): If True, no metadata will be added to
                the serialized message. Defaults to False.

        Returns:
            bytes, str: Serialized message.

        Raises:
            TypeError: If returned msg is not bytes type (str on Python 2).


        """
        if header_kwargs is None:
            header_kwargs = {}
        if isinstance(args, backwards.bytes_type) and (args == tools.YGG_MSG_EOF):
            header_kwargs['raw'] = True
        self.initialize_from_message(args, **header_kwargs)
        metadata = {'no_metadata': no_metadata}
        if add_serializer_info:
            self.debug("serializer_info = %s", str(self.serializer_info))
            metadata.update(self.serializer_info)
            metadata['typedef_base'] = self.typedef
        if header_kwargs is not None:
            metadata.update(header_kwargs)
        if header_kwargs.get('raw', False):
            data = args
        else:
            if self.func_serialize is None:
                data = args
            else:
                data = self.func_serialize(args)
                if (self.encoded_typedef['type'] == 'bytes'):
                    if not isinstance(data, backwards.bytes_type):
                        raise TypeError(("Serialization function returned object "
                                         + "of type '%s', not required '%s' type.")
                                        % (type(data), backwards.bytes_type))
                    metadata['dont_encode'] = True
                    if not no_metadata:
                        metadata['metadata'] = self.datatype.encode_type(
                            args, typedef=self.typedef)
        if ((self.initialized
             and (not tools.check_environ_bool('YGG_VALIDATE_ALL_MESSAGES')))):
            metadata.setdefault('dont_check', True)
        out = self.encoded_datatype.serialize(data, **metadata)
        return out

    def deserialize(self, msg, **kwargs):
        r"""Deserialize a message.

        Args:
            msg (str, bytes): Message to be deserialized.
            **kwargs: Additional keyword arguments are passed to the deserialize
                method of the datatype class.

        Returns:
            tuple(obj, dict): Deserialized message and header information.

        Raises:
            TypeError: If msg is not bytes type (str on Python 2).

        """
        if (((self.func_deserialize is not None)
             and (self.encoded_typedef['type'] == 'bytes'))):
            kwargs['dont_decode'] = True
        if ((self.initialized
             and (not tools.check_environ_bool('YGG_VALIDATE_ALL_MESSAGES')))):
            kwargs.setdefault('dont_check', True)
        out, metadata = self.encoded_datatype.deserialize(msg, **kwargs)
        if (self.func_deserialize is not None):
            if metadata['size'] == 0:
                out = self.empty_msg
            elif not (metadata.get('incomplete', False)
                      or metadata.get('raw', False)):
                if 'metadata' in metadata:
                    for k, v in metadata.items():
                        if k not in ['type', 'precision', 'units', 'metadata']:
                            metadata['metadata'][k] = v
                    metadata = metadata.pop('metadata')
                if not self.initialized:
                    self.update_serializer(extract=True, **metadata)
                out = self.func_deserialize(out)
        # Update serializer
        typedef_base = metadata.pop('typedef_base', {})
        typedef = copy.deepcopy(metadata)
        typedef.update(typedef_base)
        if not ((metadata.get('size', 0) == 0)
                or metadata.get('incomplete', False)
                or metadata.get('raw', False)):
            self.initialize_serializer(typedef, extract=True)
        return out, metadata

    def enable_file_header(self):  # pragma: no cover
        r"""Set serializer attributes to enable file headers to be included in
        the serializations."""
        pass

    def disable_file_header(self):
        r"""Set serializer attributes to disable file headers from being
        included in the serializations."""
        pass

    def serialize_file_header(self):  # pragma: no cover
        r"""Return the serialized header information that should be prepended
        to files serialized using this class.

        Returns:
            bytes: Header string that should be written to the file.

        """
        return b''

    def deserialize_file_header(self, fd):  # pragma: no cover
        r"""Deserialize the header information from the file and update the
        serializer.

        Args:
            fd (file): File containing header.

        """
        pass

    def consolidate_array(self, out):
        r"""Consolidate message into a structure numpy array if possible.

        Args:
            out (list, tuple, np.ndarray): Object to consolidate into a
                structured numpy array.

        Returns:
            np.ndarray: Structured numpy array containing consolidated message.

        Raises:
            ValueError: If the array cannot be consolidated.

        """
        np_dtype = self.numpy_dtype
        if np_dtype and isinstance(out, (list, tuple, np.ndarray)):
            out = serialize.consolidate_array(out, dtype=np_dtype)
        else:
            warnings.warn(("Cannot consolidate message into a structured "
                           + "numpy array: %s") % str(out))
        return out

    # def format_header(self, header_info):
    #     r"""Format header info to form a string that should prepend a message.

    #     Args:
    #         header_info (dict): Properties that should be included in the header.

    #     Returns:
    #         str: Message with header in front.

    #     """

    def parse_header(self, msg):
        r"""Extract header info from a message.

        Args:
            msg (str): Message to extract header from.

        Returns:
            dict: Message properties.

        """
        return self.datatype.deserialize(msg, no_data=True)



from yggdrasil.drivers.OutputDriver import OutputDriver


class FileOutputDriver(OutputDriver):
    r"""Class to handle output of received messages to a file.

    Args:
        name (str): Name of the output queue to receive messages from.
        args (str): Path to the file that messages should be written to.
        **kwargs: Additional keyword arguments are passed to the parent class.

    """

    _connection_type = 'file_output'
    _ocomm_type = 'FileComm'
    _schema_subtype_description = ('Connection between a model and a file.')

    def __init__(self, name, args, **kwargs):
        kwargs.setdefault('ocomm_kws', {})
        kwargs['ocomm_kws']['address'] = args
        super(FileOutputDriver, self).__init__(name, args, **kwargs)
        self.debug('(%s)', args)



from yggdrasil.tests import assert_raises, assert_equal
import yggdrasil.drivers.tests.test_ConnectionDriver as parent
from yggdrasil import runner, tools


class TestClientParam(parent.TestConnectionParam):
    r"""Test parameters for ClientDriver class."""

    def __init__(self, *args, **kwargs):
        super(TestClientParam, self).__init__(*args, **kwargs)
        self.driver = 'ClientDriver'
        self.args = None
        self.attr_list += ['comm', 'response_drivers',
                           'request_name', 'request_address']
        # Increased to allow forwarding between IPC comms on MacOS
        # self.timeout = 5.0
        self.route_timeout = 2 * self.timeout
        # self.debug_flag = True
        self.comm_name = tools.get_default_comm()
        self.server_comm = tools.get_default_comm()
        self.icomm_name = self.comm_name
        self.ocomm_name = self.server_comm

    @property
    def send_comm_kwargs(self):
        r"""dict: Keyword arguments for send comm."""
        out = self.instance.icomm.opp_comm_kwargs()
        out['comm'] = 'ClientComm'
        return out

    @property
    def recv_comm_kwargs(self):
        r"""dict: Keyword arguments for recv comm."""
        out = self.srv_drv.ocomm.opp_comm_kwargs()
        out['comm'] = 'ServerComm'
        return out

    @property
    def inst_kwargs(self):
        r"""dict: Keyword arguments for tested class."""
        out = super(TestClientParam, self).inst_kwargs
        # out['request_name'] = self.srv_drv.request_name
        out['comm'] = self.srv_drv.comm
        out['comm_address'] = self.srv_drv.comm_address
        out['icomm_kws']['comm'] = self.comm_name
        return out
    
    def setup(self, *args, **kwargs):
        r"""Recover new client message on start-up."""
        kwargs.setdefault('nprev_comm', self.comm_count)
        self.srv_drv = self.create_server()
        if not self.skip_start:
            self.srv_drv.start()
        super(TestClientParam, self).setup(*args, **kwargs)

    def teardown(self):
        r"""Recover end client message on teardown."""
        if hasattr(self, 'srv_drv'):
            self.remove_instance(self.srv_drv)
            delattr(self, 'srv_drv')
        super(TestClientParam, self).teardown()

    def create_server(self, comm_address=None):
        r"""Create a new ServerDriver instance."""
        inst = runner.create_driver(
            'ServerDriver', 'TestServerRequestDriver.' + self.uuid,
            comm=self.server_comm,
            comm_address=comm_address,
            namespace=self.namespace, working_dir=self.working_dir,
            timeout=self.timeout)
        return inst

    
class TestClientDriverNoStart(TestClientParam,
                              parent.TestConnectionDriverNoStart):
    r"""Test class for ClientDriver class without start."""

    def test_error_attributes(self):
        r"""Test error raised when trying to access attributes set on recv."""
        err_attr = ['request_id', 'model_response_address']
        for k in err_attr:
            assert_raises(AttributeError, getattr, self.instance, k)


class TestClientDriverNoInit(TestClientParam,
                             parent.TestConnectionDriverNoInit):
    r"""Test class for ClientDriver class without init."""
    pass
            

class TestClientDriver(TestClientParam, parent.TestConnectionDriver):
    r"""Test class for ClientDriver class."""

    def setup(self, *args, **kwargs):
        r"""Wait for drivers to start."""
        super(TestClientDriver, self).setup(*args, **kwargs)
        T = self.instance.start_timeout(self.timeout)
        while ((not T.is_out) and ((not self.instance.is_valid)
                                   or (not self.srv_drv.is_valid))):
            self.instance.sleep()  # pragma: debug
        self.instance.stop_timeout()

    # # Disabled so that test message is not read by mistake
    # def test_purge(self):
    #     r"""Disabled: Test purge of queue."""
    #     pass

    def test_send_recv(self, msg_send=None):
        r"""Test routing of a short message between client and server."""
        if msg_send is None:
            msg_send = self.test_msg
        T = self.instance.start_timeout(self.timeout)
        while ((not T.is_out) and ((not self.instance.is_valid)
                                   or (not self.srv_drv.is_valid))):
            self.instance.sleep()  # pragma: debug
        self.instance.stop_timeout()
        # Send a message to local output
        flag = self.send_comm.send(msg_send)
        assert(flag)
        # Receive on server side
        flag, srv_msg = self.recv_comm.recv(timeout=self.route_timeout)
        assert(flag)
        assert_equal(srv_msg, msg_send)
        self.instance.printStatus()
        self.srv_drv.printStatus()
        # Send reply back to client
        flag = self.recv_comm.send(srv_msg)
        assert(flag)
        # Receive response on client side
        flag, cli_msg = self.send_comm.recv(timeout=self.route_timeout)
        assert(flag)
        assert_equal(cli_msg, msg_send)

    def test_send_recv_nolimit(self):
        r"""Test routing of a large message between client and server."""
        self.test_send_recv(msg_send=self.msg_long)



import six
import jsonschema
from yggdrasil.metaschema.properties import MetaschemaPropertyMeta


@six.add_metaclass(MetaschemaPropertyMeta)
class MetaschemaProperty(object):
    r"""Base class for adding properties to the metaschema.

    Class Attributes:
        name (str): Name of the property. [REQUIRED]
        schema (dict): JSON schema describing valid values for the property.
            [REQUIRED]
        types (list): Types of instances that the property is valid for.
            [AUTOMATED]
        python_types (list): Python types of instances that the property is
            valid for. [AUTOMATED]

    """

    name = 'base'
    schema = None
    types = tuple()
    python_types = tuple()
    _encode = None
    _validate = None
    _compare = None
    _replaces_existing = False
    _skip_existing_validator = False
    _skip_formulaic_validator = False
    _dont_register = False

    @classmethod
    def encode(cls, instance, typedef=None):
        r"""Method to encode the property given the object.

        Args:
            instance (object): Object to get property for.
            typedef (object, None): Template value in type definition to use
                for initializing encoding some cases. Defaults to None and
                is ignored.

        Returns:
            object: Encoded property for instance.

        """
        if cls._encode is not None:
            return cls._encode(instance, typedef=typedef)
        raise NotImplementedError("Encode method not set.")
    
    @classmethod
    def validate(cls, validator, value, instance, schema):
        r"""Validator for JSON schema validation of an instance by this property.
        If there is not a user provided validate function, the instance will be
        encoded and then the encoded value will be checked against the provided
        value using cls.compare.

        Args:
            validator (jsonschmea.Validator): JSON schema validator.
            value (object): Value of the property in the schema.
            instance (object): Instance to validate.
            schema (dict): Schema that instance should be validated against.
            
        Yields:
            str: Error messages associated with failed validation.

        """
        errors = []
        if cls._validate is False:
            pass
        elif cls._validate is not None:
            errors = cls._validate(validator, value, instance, schema) or ()
        elif not cls._skip_formulaic_validator:
            is_type = False
            for t in cls.types:
                if validator.is_type(instance, t):
                    is_type = True
                    break
            if is_type:
                x = cls.encode(instance, typedef=value)
                errors = cls.compare(x, value) or ()
        for e in errors:
            yield e

    @classmethod
    def compare(cls, prop1, prop2, root1=None, root2=None):
        r"""Method to determine compatiblity of one property value with another.
        This method is not necessarily symmetric in that the second value may
        not be compatible with the first even if the first is compatible with
        the second.

        Args:
            prop1 (object): Property value to compare against prop2.
            prop2 (object): Property value to compare against.
            
        Yields:
            str: Comparision failure messages.

        """
        if cls._compare is not None:
            errors = cls._compare(prop1, prop2) or ()
            for e in errors:
                yield e
        else:
            if (prop1 != prop2):
                yield '%s is not equal to %s' % (prop1, prop2)

    @classmethod
    def normalize(cls, validator, value, instance, schema):
        r"""Method to normalize the instance based on the property value.

        Args:
            validator (Validator): Validator class.
            value (object): Property value.
            instance (object): Object to normalize.
            schema (dict): Schema containing this property.

        Returns:
            object: Normalized object.

        """
        return instance

    @classmethod
    def normalize_in_schema(cls, schema):
        r"""Method to normalize the entry for this property in a schema.

        Args:
            schema (dict): Schema to normalize.

        Returns:
            dict: Normalized schema.

        """
        return schema

    @classmethod
    def post_validate(cls, validator, value, instance, schema):
        r"""Actions performed after validation if normalizing."""
        pass

    @classmethod
    def wrapped_validate(cls, validator, value, instance, schema):
        r"""Wrapped validator that handles errors produced by the native
        validate method and ensures that the property is parsed by the base
        validator and raises the correct error if necessary.

        Args:
            *args: All arguments are passed to the validate class method.
            **kwargs: All keyword arguments are passed to the validate class
                method.

        """
        if validator._normalizing:
            validator._normalized = cls.normalize(validator, value,
                                                  validator._normalized, schema)
            instance = validator._normalized
        try:
            failed = False
            errors = cls.validate(validator, value, instance, schema) or ()
            for e in errors:
                failed = True
                yield jsonschema.ValidationError(e)
            if (((not failed) and (not cls._skip_existing_validator)
                 and (cls.name in validator._base_validator.VALIDATORS))):
                errors = validator._base_validator.VALIDATORS[cls.name](
                    validator, value, instance, schema) or ()
                for e in errors:
                    failed = True
                    yield e
        finally:
            if validator._normalizing and (not failed):
                cls.post_validate(validator, value, instance, schema)


def create_property(name, schema, encode, validate=None, compare=None,
                    dont_register=False):
    r"""Create a new property class.

    Args:
        name (str): Name of the property.
        schema (dict): JSON schema describing valid values for the property.
        encode (function): Function to encode the property based on a provided
            instance. The function must take an instance as input and return
            the value of the property for that instance.
        validate (function, optional): Function to determine if an instance
            is valid under the contraint of this property. The function must
            take as input a jsonschema validator, a property value, an
            instance to evaluate, and the schema. The function must return
            a boolean: True if the instance is valid, False otherwise. See
            cls.validate for additional information and default behavior.
        compare (function, optional): Function to determine if two property
            values are compatible. The function must take as input two
            property values and return a boolean: True if the first property
            is compatible with the second, False otherwise. See cls.compare
            for additional information and default behavior.
        dont_register (bool, optional): If True, the created property will
            not be registered. Defaults to False.

    """
    attr_dict = {'name': name, 'schema': schema,
                 '_dont_register': dont_register}
    for k, x in zip(['_encode', '_validate', '_compare'],
                    [encode, validate, compare]):
        if x is not None:
            attr_dict[k] = staticmethod(x)
    out = type(name, (MetaschemaProperty, ), attr_dict)
    return out



import os
import copy
from collections import OrderedDict
from yggdrasil import components, backwards, platform
from yggdrasil.drivers.CompiledModelDriver import (
    CompiledModelDriver, CompilerBase)
from yggdrasil.drivers.CModelDriver import CModelDriver


class MakeCompiler(CompilerBase):
    r"""Make configuration tool.

    Args:
        makefile (str, optional): Path to make file either absolute, relative to
            makedir (if provided), or relative to working_dir. Defaults to
            Makefile.
        makedir (str, optional): Directory where make should be invoked from
            if it is not the same as the directory containing the makefile.
            Defaults to directory containing makefile if provided, otherwise
            working_dir.
        target (str, optional): Make target that should be built to create the
            model executable. Defaults to None.
        **kwargs: Additional keyword arguments are passed to parent class.

    Attributes:
        makefile (str): Path to make file either relative to makedir or absolute.
        makedir (str): Directory where make should be invoked from.
        target (str): Name of executable that should be created and called.

    """
    _schema_properties = {
        'makefile': {'type': 'string', 'default': 'Makefile'},
        'makedir': {'type': 'string'},  # default will depend on makefile
        'target': {'type': 'string'}}
    toolname = 'make'
    languages = ['make', 'c', 'c++']
    platforms = ['MacOS', 'Linux']
    default_flags = ['--always-make']  # Always overwrite
    flag_options = OrderedDict([('makefile', {'key': '-f', 'position': 0})])
    output_key = ''
    no_separate_linking = True
    default_archiver = False
    linker_attributes = {'executable_ext': ''}
    
    @staticmethod
    def before_registration(cls):
        r"""Operations that should be performed to modify class attributes prior
        to registration including things like platform dependent properties and
        checking environment variables for default settings.
        """
        CompilerBase.before_registration(cls)
        if platform._is_win:  # pragma: windows
            cls.linker_attributes['executable_ext'] = '.exe'
        
    @classmethod
    def get_output_file(cls, src, target=None, **kwargs):
        r"""Determine the appropriate output file that will result when
        compiling a target.

        Args:
            src (str): Make target or source file being compiled that will be
                used to determine the path to the output file.
            target (str, optional): Target that will be used to create the
                output file instead of src if provided. Defaults to None and
                is ignored.
            **kwargs: Additional keyword arguments are passed to the parent
                class's method.

        Returns:
            str: Full path to file that will be produced.

        """
        if target is not None:
            src = target
        if src == 'clean':
            # Abort early so that working_dir not prepended
            return src
        out = super(MakeCompiler, cls).get_output_file(src, **kwargs)
        return out

    @classmethod
    def get_flags(cls, target=None, **kwargs):
        r"""Get compilation flags, replacing outfile with target.

        Args:
            target (str, optional): Target that should be built. Defaults to
                to None and is ignored.
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            list: Linker flags.

        """
        link_kws = ['linker_flags']
        for k in link_kws:
            kwargs.pop(k, None)
        kwargs.pop('outfile', None)
        # if (target is None) and (outfile is not None):
        #     target = cls.file2base(outfile)
        out = super(MakeCompiler, cls).get_flags(outfile=target, **kwargs)
        return out

    @classmethod
    def get_executable_command(cls, args, target=None, **kwargs):
        r"""Determine the command required to run the tool using the specified
        arguments and options.

        Args:
            args (list): The arguments that should be passed to the tool. If
                skip_flags is False, these are treated as input files that will
                be used by the tool.
            target (str, optional): Target that should be built. Defaults to
                to None and is set to the base name of first element in the
                provided arguments.
            **kwargs: Additional keyword arguments are passed to the parent
                class's method.

        Returns:
            str: Output to stdout from the command execution.

        """
        if not kwargs.get('skip_flags', False):
            assert(len(args) <= 1)
            if len(args) == 1:
                if target is None:
                    target = cls.file2base(args[0])
                else:
                    if target != cls.file2base(args[0]):
                        raise RuntimeError(("The argument list contents (='%s') "
                                            "and 'target' (='%s') keyword argument "
                                            "specify the same thing, but those "
                                            "provided do not match.")
                                           % (args[0], target))
            args = []
        if target is not None:
            kwargs['target'] = target
        return super(MakeCompiler, cls).get_executable_command(args, **kwargs)

    @classmethod
    def set_env(cls, logging_level=None, language=None, **kwargs):
        r"""Get environment variables that should be set for the model process.

        Args:
            logging_level (int, optional): Logging level that should be passed
                to get flags.
            language (str, optional): Language that is being compiled. Defaults
                to the first language in cls.languages.
            **kwargs: Additional keyword arguments are passed to the parent
                class's method.

        Returns:
            dict: Environment variables for the model process.

        """
        out = super(MakeCompiler, cls).set_env(**kwargs)
        if language is None:
            language = 'c'
        drv = components.import_component('model', language)
        compile_flags = drv.get_compiler_flags(
            for_model=True, logging_level=logging_level, skip_defaults=True)
        linker_flags = drv.get_linker_flags(
            for_model=True, skip_defaults=True)
        # TODO: Put these in the default_flags_env?
        # TODO: Change these to be more generic?
        out['YGGCCFLAGS'] = backwards.as_str(' '.join(compile_flags))
        out['YGGLDFLAGS'] = backwards.as_str(' '.join(linker_flags))
        # Set default compiler executable
        # compiler = drv.get_tool('compiler')
        # linker = drv.get_tool('linker')
        # if (((compiler.default_executable_env is not None)
        #      and (compiler.default_executable_env not in out))):
        #     out[compiler.default_executable_env] = compiler.get_executable()
        # if (((linker.default_executable_env is not None)
        #      and (linker.default_executable_env not in out))):
        #     out[linker.default_executable_env] = linker.get_executable()
        return out
    

class NMakeCompiler(MakeCompiler):
    toolname = 'nmake'
    platforms = ['Windows']
    default_flags = ['/NOLOGO']
    flag_options = OrderedDict([('makefile', '/f')])
    default_executable = None
    default_linker = None  # Force linker to be initialized with the same name

    @classmethod
    def language_version(cls, **kwargs):  # pragma: windows
        r"""Determine the version of this language.

        Args:
            **kwargs: Keyword arguments are passed to cls.call.

        Returns:
            str: Version of compiler/interpreter for this language.

        """
        out = cls.call(cls.version_flags, skip_flags=True,
                       allow_error=True, **kwargs)
        if 'Copyright' not in out:  # pragma: debug
            raise RuntimeError("Version call failed: %s" % out)
        return out.split('Copyright')[0]
    

class MakeModelDriver(CompiledModelDriver):
    r"""Class for running make file compiled drivers. Before running the
    make command, the necessary compiler & linker flags for the interface's
    C/C++ library are stored the environment variables YGGCCFLAGS and YGGLDFLAGS
    respectively. These should be used in the make file to correctly compile
    with the interface's C/C++ libraries.

    Args:
        name (str): Driver name.
        args (str, list): Executable that should be created (make target) and
            any arguments for the executable.
        makefile (str, optional): Path to make file either absolute, relative to
            makedir (if provided), or relative to working_dir. Defaults to
            Makefile.
        makedir (str, optional): Directory where make should be invoked from
            if it is not the same as the directory containing the makefile.
            Defaults to directory containing makefile if provided, otherwise
            working_dir.
        target (str, optional): Make target that should be built to create the
            model executable. Defaults to None.
        **kwargs: Additional keyword arguments are passed to parent class.

    Attributes:
        makefile (str): Path to make file either relative to makedir or absolute.
        makedir (str): Directory where make should be invoked from.
        target (str): Name of executable that should be created and called.

    Raises:
        RuntimeError: If neither the IPC or ZMQ C libraries are available.

    """

    _schema_subtype_description = ('Model is written in C/C++ and has a '
                                   'Makefile for compilation.')
    _schema_properties = copy.deepcopy(MakeCompiler._schema_properties)
    language = 'make'
    base_languages = ['c']

    def parse_arguments(self, args):
        r"""Sort arguments based on their syntax to determine if an argument
        is a source file, compilation flag, or runtime option/flag that should
        be passed to the model executable.

        Args:
            args (list): List of arguments provided.

        """
        # Set makedir before passing to parent class so that makedir is used
        # to normalize the model file path rather than the working directory
        # which may be different.
        if not os.path.isabs(self.makefile):
            if self.makedir is not None:
                self.makefile = os.path.normpath(
                    os.path.join(self.makedir, self.makefile))
            else:
                src_dir = os.path.dirname(args[0])
                if not os.path.isabs(src_dir):
                    src_dir = os.path.join(self.working_dir, src_dir)
                for x in [self.working_dir, src_dir]:
                    y = os.path.normpath(os.path.join(x, self.makefile))
                    if os.path.isfile(y):
                        self.makefile = y
                        break
        if self.makedir is None:
            self.makedir = os.path.dirname(self.makefile)
        kwargs = dict(default_model_dir=self.makedir)
        super(MakeModelDriver, self).parse_arguments(args, **kwargs)
        
    @classmethod
    def is_source_file(cls, fname):
        r"""Determine if the provided file name points to a source files for
        the associated programming language by checking the extension.

        Args:
            fname (str): Path to file.

        Returns:
            bool: True if the provided file is a source file, False otherwise.

        """
        compiler = cls.get_tool('compiler')
        for lang in compiler.languages:
            if lang == cls.language:
                continue
            drv = components.import_component('model', lang)
            if drv.is_source_file(fname):
                return True
        return False
        
    def compile_model(self, target=None, **kwargs):
        r"""Compile model executable(s).

        Args:
            target (str, optional): Target to build.
            **kwargs: Keyword arguments are passed on to the parent class's
                method.

        """
        if target is None:
            target = self.target
        if target == 'clean':
            return self.call_compiler([], target=target,
                                      out=target, overwrite=True,
                                      makefile=self.makefile,
                                      working_dir=self.working_dir, **kwargs)
        else:
            default_kwargs = dict(skip_interface_flags=True,
                                  # source_files=[],  # Unknown source files, use target
                                  for_model=False,  # flags are in environment
                                  working_dir=self.makedir,
                                  makefile=self.makefile,
                                  env=self.set_env(for_compile=True))
            if target is not None:
                default_kwargs['target'] = target
            for k, v in default_kwargs.items():
                kwargs.setdefault(k, v)
            return super(MakeModelDriver, self).compile_model(**kwargs)
        
    def set_env(self, for_compile=False):
        r"""Get environment variables that should be set for the model process.

        Args:
            for_compile (bool, optional): If True, environment variables are set
                that are necessary for compiling. Defaults to False.

        Returns:
            dict: Environment variables for the model process.

        """
        out = super(MakeModelDriver, self).set_env(for_compile=for_compile)
        if not for_compile:
            # TODO: Pass language?
            out = CModelDriver.update_ld_library_path(out)
        return out
        
    def cleanup(self):
        r"""Remove compiled executable."""
        if (self.model_file is not None) and os.path.isfile(self.model_file):
            self.compile_model(target='clean')
        super(MakeModelDriver, self).cleanup()



from yggdrasil.metaschema.datatypes import MetaschemaTypeError
from yggdrasil.metaschema.datatypes.MetaschemaType import MetaschemaType
from yggdrasil.metaschema.datatypes.JSONObjectMetaschemaType import (
    JSONObjectMetaschemaType)
from yggdrasil.metaschema.properties.ArgsMetaschemaProperty import (
    ArgsMetaschemaProperty)


class InstanceMetaschemaType(MetaschemaType):
    r"""Type for evaluating instances of Python classes."""

    name = 'instance'
    description = 'Type for Python class instances.'
    properties = ['class', 'args']
    definition_properties = ['class']
    metadata_properties = ['class', 'args']
    extract_properties = ['class', 'args']
    python_types = (object, )
    cross_language_support = False

    @classmethod
    def validate(cls, obj, raise_errors=False):
        r"""Validate an object to check if it could be of this type.

        Args:
            obj (object): Object to validate.
            raise_errors (bool, optional): If True, errors will be raised when
                the object fails to be validated. Defaults to False.

        Returns:
            bool: True if the object could be of this type, False otherwise.

        """
        # Base not called because every python object should pass validation
        # against the object class
        try:
            ArgsMetaschemaProperty.instance2args(obj)
            return True
        except MetaschemaTypeError:
            if raise_errors:
                raise ValueError("Class dosn't have an input_args attribute.")
            return False
        
    @classmethod
    def encode_data(cls, obj, typedef):
        r"""Encode an object's data.

        Args:
            obj (object): Object to encode.
            typedef (dict): Type definition that should be used to encode the
                object.

        Returns:
            string: Encoded object.

        """
        args = ArgsMetaschemaProperty.instance2args(obj)
        if isinstance(typedef, dict) and ('args' in typedef):
            typedef_args = {'properties': typedef['args']}
        else:
            typedef_args = None
        return JSONObjectMetaschemaType.encode_data(args, typedef_args)

    @classmethod
    def decode_data(cls, obj, typedef):
        r"""Decode an object.

        Args:
            obj (string): Encoded object to decode.
            typedef (dict): Type definition that should be used to decode the
                object.

        Returns:
            object: Decoded object.

        """
        # TODO: Normalization can be removed if metadata is normalized
        typedef = cls.normalize_definition(typedef)
        args = JSONObjectMetaschemaType.decode_data(
            obj, {'properties': typedef.get('args', {})})
        return typedef['class'](**args)



import os
from yggdrasil.examples.tests import TestExample


class TestExampleFakeplant(TestExample):
    r"""Test the Fakeplant example."""

    example_name = 'fakeplant'

    @property
    def output_dir(self):
        r"""str: Output directory."""
        if self.yamldir is None:  # pragma: debug
            return None
        return os.path.join(self.yamldir, 'Output')

    @property
    def output_files(self):
        r"""list: Output files for the run."""
        outdir = self.output_dir
        filelist = ['canopy_structure.txt',
                    'growth_rate.txt',
                    'light_intensity.txt',
                    'photosynthesis_rate.txt']
        out = [os.path.join(outdir, f) for f in filelist]
        return out

    def check_results(self):
        r"""This should be overridden with checks for the result."""
        pass

    def run_example(self):
        r"""This runs an example in the correct language."""
        if self.output_dir is not None:
            if not os.path.isdir(self.output_dir):
                os.mkdir(self.output_dir)
        super(TestExampleFakeplant, self).run_example()



from yggdrasil.drivers.ClientDriver import ClientDriver


class RMQAsyncClientDriver(ClientDriver):
    r"""Class for handling an RMQAsync client.

    Args:
        *args: Arguments are passed to parent class.
        **kwargs: Keyword arguments are passed to parent class.

    """

    _connection_type = 'rmq_async_client'
    _schema_subtype_description = ('Connection between a model acting as a '
                                   'client and an asynchronous RabbitMQ server '
                                   'request comm.')
    
    def __init__(self, *args, **kwargs):
        kwargs['comm'] = 'RMQAsyncComm'
        super(RMQAsyncClientDriver, self).__init__(*args, **kwargs)
        self.timeout = 10.0



import os
from yggdrasil.config import ygg_cfg
from yggdrasil import tools


class Driver(tools.YggThreadLoop):
    r"""Base class for all drivers.

    Args:
        name (str): Driver name.
        yml (dict, optional): Dictionary of yaml specification options for this
            driver. Defaults to empty dict.
        env (dict, optional): Dictionary of environment variables that should
            be set when the driver starts. Defaults to {}.
        comm_env (dict, optional): Dictionary of environment variables for
            paired IO communication drivers. Defaults to {}.
        namespace (str, optional): Namespace for set of drivers running
            together. If not provided, the config option ('rmq', 'namespace')
            is used.
        rank (int, optional): Rank of the integration. Defaults to None.

    Attributes:
        name (str): Driver name.
        env (dict): Dictionary of environment variables.
        comm_env (dict): Dictionary of environment variables for paired IO
            communication drivers.
        yml (dict): Dictionary of yaml specification options for this driver.
        namespace (str): Namespace for set of drivers running together.
        rank (int): Rank of the integration.

    """
    # =========================================================================
    # METHODS THAT MUST HAVE SUPER AT BEGINNING AND CAN BE OVERRIDEN BY CHILD
    # CLASSES TO ADD DRIVER FUNCTIONALITY. ALL OF THE CHILD CLASSES MUST HAVE
    # COMPATIBLE FORMATS (THE SAME NAMED ARGUMENTS).
    def __init__(self, name, yml=None, env=None, comm_env=None, namespace=None,
                 rank=None, **kwargs):
        if yml is None:
            yml = {}
        if env is None:
            env = {}
        if comm_env is None:
            comm_env = {}
        # Check if thread initialized to avoid doing it twice for drivers
        # with multiple inheritance that both need to call __init__
        if getattr(self, '_thread_initialized', False):  # pragma: debug
            raise Exception("Thread already initialized. "
                            + "Check multiple inheritance")
        super(Driver, self).__init__(name, **kwargs)
        self._thread_initialized = True
        self.debug('')
        # if ygg_cfg.get('debug', 'ygg') == 'DEBUG':
        #     self.sleeptime = 1.0
        # Set defaults
        if namespace is None:
            namespace = ygg_cfg.get('rmq', 'namespace')
            self.debug("Setting namespace to %s", namespace)
        if kwargs.get('working_dir', None) is None:
            if isinstance(yml, dict) and ('working_dir' in yml):
                self.working_dir = yml['working_dir']
            else:
                self.working_dir = os.getcwd()
        # Assign things
        self.yml = yml
        self.env = env
        self.comm_env = comm_env
        self.namespace = namespace
        self.rank = rank
        self._term_meth = "terminate"

    @property
    def is_valid(self):
        r"""bool: True if the driver is functional."""
        return True

    def stop(self):
        r"""Stop the driver."""
        if self.was_terminated:
            self.debug('Driver already terminated.')
            return
        self.debug('')
        self._term_meth = 'stop'
        self.graceful_stop()
        self.terminate()

    def graceful_stop(self):
        r"""Gracefully stop the driver."""
        self.debug('')

    def do_terminate(self):
        r"""Actions that should stop the driver."""
        self.debug('Returning')

    def terminate(self):
        r"""Stop the driver, without attempting to allow it to finish."""
        if self.was_terminated:
            self.debug('Driver already terminated.')
            return
        self.do_terminate()
        self.debug('')
        super(Driver, self).terminate()
        self.debug('Returning')

    def on_model_exit(self):
        r"""Processes that should be run when an associated model exits."""
        self.debug('')



r"""IO and Model drivers."""
from yggdrasil.components import import_component


_non_component_modules = ['lpy_model.py']


def create_driver(driver=None, name=None, args=None, **kwargs):
    r"""Dynamically create a driver based on a string and other driver
    properties.

    Args:
        driver (str): Name of the driver that should be created.
        name (str): Name to give the driver.
        args (object, optional): Second argument for drivers which take a
            minimum of two arguments. If None, the driver is assumed to take a
            minimum of one argument. Defaults to None.
        **kwargs: Additional keyword arguments are passed to the driver
            class.

    Returns:
        object: Instance of the requested driver.

    """
    class_ = import_component('model', driver, without_schema=True)
    if args is None:
        instance = class_(name, **kwargs)
    else:
        instance = class_(name, args, **kwargs)
    return instance


__all__ = ['create_driver', 'Driver',
           'ModelDriver', 'PythonModelDriver',
           'CModelDriver', 'CPPModelDriver',
           'MakeModelDriver', 'MatlabModelDriver', 'LPyModelDriver',
           'ConnectionDriver', 'InputDriver', 'OutputDriver',
           'FileInputDriver', 'FileOutputDriver',
           'ClientDriver', 'ServerDriver',
           'RMQInputDriver', 'RMQOutputDriver',
           'RMQClientDriver', 'RMQServerDriver']



import glob
import os
from yggdrasil.examples.tests import TestExample


class TestExampleModelError(TestExample):
    r"""Test the model_error example."""

    example_name = 'model_error'
    expects_error = True

    @property
    def core_dump(self):
        r"""str: Pattern for core dump that may be produced."""
        if self.yamldir is None:  # pragma: no cover
            return None
        return os.path.join(self.yamldir, 'core.*')

    def cleanup(self):
        r"""Cleanup files created during the test."""
        super(TestExampleModelError, self).cleanup()
        if self.core_dump is not None:
            fcore = glob.glob(self.core_dump)
            for f in fcore:  # pragma: debug
                os.remove(f)



from yggdrasil import scanf, backwards
from yggdrasil.tests import assert_equal


def test_scanf():
    r"""Test scanf."""
    num_codes = 'dieEfgGouxX'
    fmt_list = [("%g%+gj", complex(1)),
                ("%s", "hello"),
                ("%5s", "the"),
                ("%5s\t%ld\t%lf\t%lf%+lfj\n", ("one", 1, 1.0, complex(1, 1)))]
    fmt_list += [("%" + s, 1) for s in num_codes]
    fmt_list += [("%5.2" + s, 1) for s in num_codes]
    fmt_list += [("%+5.2" + s, 1) for s in num_codes]
    fmt_list += [("%-5.2" + s, 1) for s in num_codes]
    fmt_list += [("% 5.2" + s, 1) for s in num_codes]
    fmt_list += [("%05.2" + s, 1) for s in num_codes]
    for fmt, val in fmt_list:
        if isinstance(val, tuple):
            val_tup = val
        else:
            val_tup = (val, )
        new_tup = []
        for v in val_tup:
            if isinstance(v, complex):
                new_tup += [v.real, v.imag]
            else:
                new_tup.append(v)
        val_str = fmt % tuple(new_tup)
        res = scanf.scanf(fmt, val_str)
        assert_equal(res, val_tup)
        # Test bytes version
        res = backwards.scanf_bytes(fmt, backwards.as_bytes(val_str))
        assert_equal(res, backwards.as_bytes(val_tup, recurse=True,
                                             allow_pass=True))
        # Test unicode version
        res = backwards.scanf_bytes(fmt, backwards.as_unicode(val_str))
        assert_equal(res, backwards.as_unicode(val_tup, recurse=True,
                                               allow_pass=True))



from __future__ import print_function
from yggdrasil.interface.YggInterface import YggInput, YggOutput


def runhello():
    print('Hello from Python')

    # Ins/outs matching with the the model yaml
    inf = YggInput('inFile')
    outf = YggOutput('outFile')
    inq = YggInput('helloQueueIn')
    outq = YggOutput('helloQueueOut')
    print("hello(P): Created I/O channels")

    # Receive input from a local file
    ret, buf = inf.recv()
    if not ret:
        raise RuntimeError('hello(P): ERROR FILE RECV')
    print('hello(P): Received %d bytes from file: %s' % (len(buf), buf))

    # Send output to the output queue
    ret = outq.send(buf)
    if not ret:
        raise RuntimeError('hello(P): ERROR QUEUE SEND')
    print('hello(P): Sent to outq')

    # Receive input form the input queue
    ret, buf = inq.recv()
    if not ret:
        raise RuntimeError('hello(P): ERROR QUEUE RECV')
    print('hello(P): Received %d bytes from queue: %s' % (len(buf), buf))

    # Send output to a local file
    ret = outf.send(buf)
    if not ret:
        raise RuntimeError('hello(P): ERROR FILE SEND')
    print('hello(P): Sent to outf')

    print('Goodbye from Python')

    
if __name__ == '__main__':
    runhello()



import unittest
from yggdrasil.communication.RMQComm import RMQComm
import yggdrasil.drivers.tests.test_ClientDriver as parent


_rmq_installed = RMQComm.is_installed(language='python')


@unittest.skipIf(not _rmq_installed, "RMQ Server not running")
class TestRMQClientParam(parent.TestClientParam):
    r"""Test parameters for RMQClientDriver class."""

    def __init__(self, *args, **kwargs):
        super(TestRMQClientParam, self).__init__(*args, **kwargs)
        self.driver = 'RMQClientDriver'
        self.server_comm = 'RMQComm'
        self.ocomm_name = self.server_comm

    
@unittest.skipIf(not _rmq_installed, "RMQ Server not running")
class TestRMQClientDriverNoStart(TestRMQClientParam,
                                 parent.TestClientDriverNoStart):
    r"""Test class for RMQClientDriver class without start."""
    pass


@unittest.skipIf(not _rmq_installed, "RMQ Server not running")
class TestRMQClientDriver(TestRMQClientParam, parent.TestClientDriver):
    r"""Test class for RMQClientDriver class."""
    pass



import uuid
from yggdrasil.components import import_component
from yggdrasil.communication import (CommBase, new_comm, get_comm)


class ClientComm(CommBase.CommBase):
    r"""Class for handling Client side communication.

    Args:
        name (str): The environment variable where communication address is
            stored.
        request_comm (str, optional): Comm class that should be used for the
            request comm. Defaults to None.
        response_kwargs (dict, optional): Keyword arguments for the response
            comm. Defaults to empty dict.
        **kwargs: Additional keywords arguments are passed to the output comm.

    Attributes:
        response_kwargs (dict): Keyword arguments for the response comm.
        icomm (dict): Response comms keyed to the ID of the associated request.
        icomm_order (list): Response comm keys in the order or the requests.
        ocomm (Comm): Request comm.

    """

    _dont_register = True
    
    def __init__(self, name, request_comm=None, response_kwargs=None,
                 dont_open=False, **kwargs):
        if response_kwargs is None:
            response_kwargs = dict()
        ocomm_name = name
        ocomm_kwargs = kwargs
        ocomm_kwargs['direction'] = 'send'
        ocomm_kwargs['dont_open'] = True
        ocomm_kwargs['comm'] = request_comm
        self.response_kwargs = response_kwargs
        self.ocomm = get_comm(ocomm_name, **ocomm_kwargs)
        self.icomm = dict()
        self.icomm_order = []
        self.response_kwargs.setdefault('comm', self.ocomm.comm_class)
        self.response_kwargs.setdefault('recv_timeout', self.ocomm.recv_timeout)
        super(ClientComm, self).__init__(self.ocomm.name, dont_open=dont_open,
                                         recv_timeout=self.ocomm.recv_timeout,
                                         is_interface=self.ocomm.is_interface,
                                         direction='send', no_suffix=True,
                                         address=self.ocomm.address)

    @classmethod
    def is_installed(cls, language=None):
        r"""Determine if the necessary libraries are installed for this
        communication class.

        Args:
            language (str, optional): Specific language that should be checked
                for compatibility. Defaults to None and all languages supported
                on the current platform will be checked.

        Returns:
            bool: Is the comm installed.

        """
        return import_component('comm').is_installed(language=language)

    @property
    def maxMsgSize(self):
        r"""int: Maximum size of a single message that should be sent."""
        return self.ocomm.maxMsgSize
        
    @classmethod
    def underlying_comm_class(self):
        r"""str: Name of underlying communication class."""
        return import_component('comm').underlying_comm_class()

    @classmethod
    def comm_count(cls):
        r"""int: Number of communication connections."""
        return import_component('comm').comm_count()

    @classmethod
    def new_comm_kwargs(cls, name, request_comm=None, **kwargs):
        r"""Initialize communication with new comms.

        Args:
            name (str): Name for new comm.
            request_comm (str, optional): Name of class for new output comm.
                Defaults to None.

        """
        args = [name]
        ocomm_class = import_component('comm', request_comm)
        kwargs['direction'] = 'send'
        if 'address' not in kwargs:
            oargs, kwargs = ocomm_class.new_comm_kwargs(name, **kwargs)
        kwargs['request_comm'] = request_comm
        return args, kwargs

    @property
    def opp_comms(self):
        r"""dict: Name/address pairs for opposite comms."""
        out = super(ClientComm, self).opp_comms
        out.update(**self.ocomm.opp_comms)
        return out

    def opp_comm_kwargs(self):
        r"""Get keyword arguments to initialize communication with opposite
        comm object.

        Returns:
            dict: Keyword arguments for opposite comm object.

        """
        kwargs = super(ClientComm, self).opp_comm_kwargs()
        kwargs['comm'] = "ServerComm"
        kwargs['request_comm'] = self.ocomm.comm_class
        kwargs['response_kwargs'] = self.response_kwargs
        return kwargs
        
    def open(self):
        r"""Open the connection."""
        super(ClientComm, self).open()
        self.ocomm.open()

    def close(self, *args, **kwargs):
        r"""Close the connection."""
        self.ocomm.close(*args, **kwargs)
        for k in self.icomm_order:
            self.icomm[k].close()
        super(ClientComm, self).close(*args, **kwargs)

    @property
    def is_open(self):
        r"""bool: True if the connection is open."""
        return self.ocomm.is_open

    @property
    def is_closed(self):
        r"""bool: True if the connection is closed."""
        return self.ocomm.is_closed

    @property
    def n_msg_send(self):
        r"""int: The number of messages in the connection."""
        return self.ocomm.n_msg_send

    @property
    def n_msg_send_drain(self):
        r"""int: The number of outgoing messages in the connection to drain."""
        return self.ocomm.n_msg_send_drain

    # RESPONSE COMM
    def create_response_comm(self):
        r"""Create a response comm based on information from the last header."""
        comm_kwargs = dict(direction='recv', is_response_client=True,
                           single_use=True, **self.response_kwargs)
        header = dict(request_id=str(uuid.uuid4()))
        if header['request_id'] in self.icomm:  # pragma: debug
            raise ValueError("Request ID %s already in use." % header['request_id'])
        c = new_comm('client_response_comm.' + header['request_id'], **comm_kwargs)
        header['response_address'] = c.address
        self.icomm[header['request_id']] = c
        self.icomm_order.append(header['request_id'])
        return header

    def remove_response_comm(self):
        r"""Remove response comm."""
        key = self.icomm_order.pop(0)
        icomm = self.icomm.pop(key)
        icomm.close()

    # SEND METHODS
    def send(self, *args, **kwargs):
        r"""Create a response comm and then send a message to the output comm
        with the response address in the header.

        Args:
            *args: Arguments are passed to output comm send method.
            **kwargs: Keyword arguments are passed to output comm send method.

        Returns:
            obj: Output from output comm send method.

        """
        msg = args[0]
        # if self.is_closed:
        #     self.debug("send(): Connection closed.")
        #     return False
        if msg != self.eof_msg:
            kwargs['header_kwargs'] = self.create_response_comm()
        out = self.ocomm.send(*args, **kwargs)
        if (not out) and (msg != self.eof_msg):
            self.remove_response_comm()
        return out

    # RECV METHODS
    def recv(self, *args, **kwargs):
        r"""Receive a message from the input comm and open a new response comm
        for output using address from the header.

        Args:
            *args: Arguments are passed to input comm recv method.
            **kwargs: Keyword arguments are passed to input comm recv method.

        Returns:
            obj: Output from input comm recv method.

        """
        # if self.is_closed:
        #     self.debug("recv(): Connection closed.")
        #     return (False, None)
        if len(self.icomm) == 0:  # pragma: debug
            raise RuntimeError("There are not any registered response comms.")
        out = self.icomm[self.icomm_order[0]].recv(*args, **kwargs)
        self.remove_response_comm()
        return out

    # CALL
    def call(self, *args, **kwargs):
        r"""Do RPC call. The request message is sent to the output comm and the
        response is received from the input comm.

        Args:
            *args: Arguments are passed to output comm send method.
            **kwargs: Keyword arguments are passed to output comm send method

        Returns:
            obj: Output from input comm recv method.

        """
        flag = self.send(*args, **kwargs)
        if not flag:  # pragma: debug
            return (False, self.empty_obj_recv)
        return self.recv(timeout=False)

    def call_nolimit(self, *args, **kwargs):
        r"""Alias for call."""
        return self.call(*args, **kwargs)

    # OLD STYLE ALIASES
    def rpcSend(self, *args, **kwargs):
        r"""Alias for RPCComm.send"""
        return self.send(*args, **kwargs)

    def rpcRecv(self, *args, **kwargs):
        r"""Alias for RPCComm.recv"""
        return self.recv(*args, **kwargs)
    
    def rpcCall(self, *args, **kwargs):
        r"""Alias for RPCComm.call"""
        return self.call(*args, **kwargs)
    
    def drain_messages(self, direction='send', **kwargs):
        r"""Sleep while waiting for messages to be drained."""
        if direction == 'send':
            self.ocomm.drain_messages(direction='send', **kwargs)

    # def purge(self):
    #     r"""Purge input and output comms."""
    #     self.ocomm.purge()
    #     # Unsure if client should purge all input comms...
    #     # for k in self.icomm_order:
    #     #     self.icomm[k].purge()
    #     super(ClientComm, self).purge()



import unittest
from yggdrasil.communication.RMQComm import RMQComm
import yggdrasil.drivers.tests.test_ServerDriver as parent


_rmq_installed = RMQComm.is_installed(language='python')


@unittest.skipIf(not _rmq_installed, "RMQ Server not running")
class TestRMQServerParam(parent.TestServerParam):
    r"""Test parameters for RMQServerDriver class."""

    def __init__(self, *args, **kwargs):
        super(TestRMQServerParam, self).__init__(*args, **kwargs)
        self.driver = 'RMQServerDriver'
        self.client_comm = 'RMQComm'
        self.icomm_name = self.client_comm

    
@unittest.skipIf(not _rmq_installed, "RMQ Server not running")
class TestRMQServerDriverNoStart(TestRMQServerParam,
                                 parent.TestServerDriverNoStart):
    r"""Test class for RMQServerDriver class without start."""
    pass


@unittest.skipIf(not _rmq_installed, "RMQ Server not running")
class TestRMQServerDriver(TestRMQServerParam, parent.TestServerDriver):
    r"""Test class for RMQServerDriver class."""
    pass



import os
import copy
import numpy as np
import warnings
from yggdrasil import backwards
from yggdrasil.metaschema.encoder import encode_json, decode_json
from yggdrasil.metaschema.datatypes import _schema_dir
from yggdrasil.metaschema.datatypes.JSONObjectMetaschemaType import (
    JSONObjectMetaschemaType)
from yggdrasil.metaschema.datatypes.PlyMetaschemaType import (
    PlyDict,
    _index_type, _color_type, _coord_type,
    _index_conv, _color_conv, _coord_conv,
    _index_fmt, _color_fmt, _coord_fmt)
    

_schema_file = os.path.join(_schema_dir, 'obj.json')
_default_element_order = ['material', 'vertices', 'params', 'normals', 'texcoords',
                          'points', 'lines', 'faces', 'curves', 'curve2Ds', 'surfaces']
# TODO: Unclear what standard puts colors after coords and how that is
# reconciled with the weight (i.e. do colors go before or after weight)
_default_property_order = {
    'vertices': ['x', 'y', 'z', 'red', 'green', 'blue', 'w'],
    'params': ['u', 'v', 'w'],
    'normals': ['i', 'j', 'k'],
    'texcoords': ['u', 'v', 'w'],
    'points': 'vertex_indices',
    'lines': ['vertex_index', 'texcoord_index'],
    'faces': ['vertex_index', 'texcoord_index', 'normal_index'],
    'curves': ['starting_param', 'ending_param', ['vertex_indices']],
    'curve2Ds': 'param_indices',
    'surfaces': ['starting_param_u', 'ending_param_u',
                 'starting_param_v', 'ending_param_v',
                 {'vertex_indices': ['vertex_index', 'texcoord_index', 'normal_index']}]}
_index_properties = ['vertex_indices', 'vertex_index', 'texcoord_index',
                     'normal_index', 'param_indices']
_default_property_formats = {}
_default_property_converters = {}
for k in ['x', 'y', 'z', 'u', 'v', 'w', 'i', 'j', 'k',
          'starting_param', 'ending_param',
          'starting_param_u', 'ending_param_u',
          'starting_param_v', 'ending_param_v']:
    _default_property_formats[k] = _coord_fmt
    _default_property_converters[k] = _coord_conv
for k in ['red', 'green', 'blue']:
    _default_property_formats[k] = _color_fmt
    _default_property_converters[k] = _color_conv
for k in ['vertex_index', 'texcoord_index', 'normal_index', 'param_index',
          'vertex_indices', 'param_indices']:
    _default_property_formats[k] = _index_fmt
    _default_property_converters[k] = _index_conv
_map_element2code = {'material': 'usemtl', 'vertices': 'v',
                     'params': 'vp', 'normals': 'vn', 'texcoords': 'vt',
                     'points': 'p', 'lines': 'l', 'faces': 'f',
                     'curves': 'curv', 'curve2Ds': 'curv2', 'surfaces': 'surf'}
_map_code2element = {v: k for k, v in _map_element2code.items()}


def create_schema(overwrite=False):
    r"""Creates a file containing the Obj schema.

    Args:
        overwrite (bool, optional): If True and a file already exists, the
            existing file will be replaced. If False, an error will be raised
            if the file already exists.

    """
    if (not overwrite) and os.path.isfile(_schema_file):
        raise RuntimeError("Schema file already exists.")
    schema = {
        'title': 'obj',
        'description': 'A mapping container for Obj 3D data.',
        'type': 'object',
        'required': ['vertices', 'faces'],
        'definitions': {
            'vertex': {
                'description': 'Map describing a single vertex.',
                'type': 'object', 'required': ['x', 'y', 'z'],
                'additionalProperties': False,
                'properties': {'x': {'type': _coord_type},
                               'y': {'type': _coord_type},
                               'z': {'type': _coord_type},
                               'red': {'type': _color_type},
                               'blue': {'type': _color_type},
                               'green': {'type': _color_type},
                               'w': {'type': _coord_type, 'default': 1.0}}},
            'param': {
                'description': 'Map describing a single parameter space point.',
                'type': 'object', 'required': ['u', 'v'],
                'additionalProperties': False,
                'properties': {'u': {'type': _coord_type},
                               'v': {'type': _coord_type},
                               'w': {'type': _coord_type, 'default': 1.0}}},
            'normal': {
                'description': 'Map describing a single normal.',
                'type': 'object', 'required': ['i', 'j', 'k'],
                'additionalProperties': False,
                'properties': {'i': {'type': _coord_type},
                               'j': {'type': _coord_type},
                               'k': {'type': _coord_type}}},
            'texcoord': {
                'description': 'Map describing a single texture vertex.',
                'type': 'object', 'required': ['u'],
                'additionalProperties': False,
                'properties': {'u': {'type': _coord_type},
                               'v': {'type': _coord_type, 'default': 0.0},
                               'w': {'type': _coord_type, 'default': 0.0}}},
            'point': {
                'description': 'Array of vertex indices describing a set of points.',
                'type': 'array', 'minItems': 1,
                'items': {'type': _index_type}},
            'line': {
                'description': ('Array of vertex indices and texture indices '
                                + 'describing a line.'),
                'type': 'array', 'minItems': 2,
                'items': {'type': 'object', 'required': ['vertex_index'],
                          'additionalProperties': False,
                          'properties':
                              {'vertex_index': {'type': _index_type},
                               'texcoord_index': {'type': _index_type}}}},
            'face': {
                'description': ('Array of vertex, texture, and normal indices '
                                + 'describing a face.'),
                'type': 'array', 'minItems': 3,
                'items': {'type': 'object', 'required': ['vertex_index'],
                          'additionalProperties': False,
                          'properties':
                              {'vertex_index': {'type': _index_type},
                               'texcoord_index': {'type': _index_type},
                               'normal_index': {'type': _index_type}}}},
            'curve': {
                'description': 'Properties of describing a curve.',
                'type': 'object', 'required': ['starting_param', 'ending_param',
                                               'vertex_indices'],
                'additionalProperties': False,
                'properties': {
                    'starting_param': {'type': _coord_type},
                    'ending_param': {'type': _coord_type},
                    'vertex_indices': {
                        'type': 'array', 'minItems': 2,
                        'items': {'type': _index_type}}}},
            'curve2D': {
                'description': ('Array of parameter indices describine a 2D curve on '
                                + 'a surface.'),
                'type': 'array', 'minItems': 2,
                'items': {'type': _index_type}},
            'surface': {
                'description': 'Properties describing a surface.',
                'type': 'object', 'required': ['starting_param_u', 'ending_param_u',
                                               'starting_param_v', 'ending_param_v',
                                               'vertex_indices'],
                'additionalProperties': False,
                'properties': {
                    'starting_param_u': {'type': _coord_type},
                    'ending_param_u': {'type': _coord_type},
                    'starting_param_v': {'type': _coord_type},
                    'ending_param_v': {'type': _coord_type},
                    'vertex_indices': {
                        'type': 'array', 'minItems': 2,
                        'items': {'type': 'object', 'required': ['vertex_index'],
                                  'additionalProperties': False,
                                  'properties': {
                                      'vertex_index': {'type': _index_type},
                                      'texcoord_index': {'type': _index_type},
                                      'normal_index': {'type': _index_type}}}}}}},
        'properties': {
            'material': {
                'description': 'Name of the material to use.',
                'type': ['unicode', 'string']},
            'vertices': {
                'description': 'Array of vertices.',
                'type': 'array', 'items': {'$ref': '#/definitions/vertex'}},
            'params': {
                'description': 'Array of parameter coordinates.',
                'type': 'array', 'items': {'$ref': '#/definitions/param'}},
            'normals': {
                'description': 'Array of normals.',
                'type': 'array', 'items': {'$ref': '#/definitions/normal'}},
            'texcoords': {
                'description': 'Array of texture vertices.',
                'type': 'array', 'items': {'$ref': '#/definitions/texcoord'}},
            'points': {
                'description': 'Array of points.',
                'type': 'array', 'items': {'$ref': '#/definitions/point'}},
            'lines': {
                'description': 'Array of lines.',
                'type': 'array', 'items': {'$ref': '#/definitions/line'}},
            'faces': {
                'description': 'Array of faces.',
                'type': 'array', 'items': {'$ref': '#/definitions/face'}},
            'curves': {
                'description': 'Array of curves.',
                'type': 'array', 'items': {'$ref': '#/definitions/curve'}},
            'curve2Ds': {
                'description': 'Array of curve2Ds.',
                'type': 'array', 'items': {'$ref': '#/definitions/curve2D'}},
            'surfaces': {
                'description': 'Array of surfaces.',
                'type': 'array', 'items': {'$ref': '#/definitions/surface'}}},
        'dependencies': {
            'lines': ['vertices'],
            'faces': ['vertices'],
            'curves': ['vertices'],
            'curve2Ds': ['params'],
            'surfaces': ['vertices']}}
    with open(_schema_file, 'w') as fd:
        encode_json(schema, fd, indent='\t')


def get_schema():
    r"""Return the Obj schema, initializing it if necessary.

    Returns:
        dict: Obj schema.
    
    """
    if not os.path.isfile(_schema_file):
        create_schema()
    with open(_schema_file, 'r') as fd:
        out = decode_json(fd)
    return out


if not os.path.isfile(_schema_file):  # pragma: debug
    create_schema()


class ObjDict(PlyDict):
    r"""Enhanced dictionary class for storing Obj information."""

    @property
    def mesh(self):
        r"""list: Vertices for each face in the structure."""
        mesh = []
        for i in range(self.count_elements('faces')):
            imesh = []
            for f in self['faces']:
                for v in f:
                    imesh += [self['vertices'][v['vertex_index']][k]
                              for k in ['x', 'y', 'z']]
            mesh.append(imesh)
        return mesh

    @classmethod
    def from_shape(cls, shape, d, conversion=1.0):  # pragma: lpy
        r"""Create a ply dictionary from a PlantGL shape and descritizer.

        Args:
            scene (openalea.plantgl.scene): Scene that should be descritized.
            d (openalea.plantgl.descritizer): Descritizer.
            conversion (float, optional): Conversion factor that should be
                applied to the vertex positions. Defaults to 1.0.

        """
        iobj = super(ObjDict, cls).from_shape(shape, d, conversion=conversion,
                                              _as_obj=True)
        if iobj is not None:
            # Texcoords
            if d.result.texCoordList:
                iobj.setdefault('texcoords', [])
                for t in d.result.texCoordList:
                    # TODO: Should the coords be scaled?
                    iobj['texcoords'].append({'u': t.x, 'v': t.y})
                if d.result.texCoordIndexList:
                    for i, t in enumerate(d.result.texCoordIndexList):
                        if t[0] < len(iobj['texcoords']):
                            for j in range(3):
                                iobj['faces'][i][j]['texcoord_index'] = t[j]
            # Normals
            if d.result.normalList:
                iobj.setdefault('normals', [])
                for n in d.result.normalList:
                    iobj['normals'].append({'i': n.x, 'j': n.y, 'k': n.z})
                if d.result.normalIndexList:
                    for i, n in enumerate(d.result.normalIndexList):
                        if n[0] < len(iobj['normals']):
                            for j in range(3):
                                iobj['faces'][i][j]['normal_index'] = n[j]
        return iobj

    def to_geom_args(self, conversion=1.0, name=None):  # pragma: lpy
        r"""Get arguments for creating a PlantGL geometry.

        Args:
            conversion (float, optional): Conversion factor that should be
                applied to the vertices. Defaults to 1.0.
            name (str, optional): Name that should be given to the created
                PlantGL symbol. Defaults to None and is ignored.

        Returns:
            tuple: Class, arguments and keyword arguments for PlantGL geometry.

        """
        import openalea.plantgl.all as pgl
        smb_class, args, kwargs = super(ObjDict, self).to_geom_args(
            conversion=conversion, name=name, _as_obj=True)
        index_class = pgl.Index
        array_class = pgl.IndexArray
        # Texture coords
        if self.get('texcoords', []):
            obj_texcoords = []
            for t in self['texcoords']:
                obj_texcoords.append(pgl.Vector2(np.float64(t['u']),
                                                 np.float64(t.get('v', 0.0))))
            kwargs['texCoordList'] = pgl.Point2Array(obj_texcoords)
            obj_ftexcoords = []
            for i, f in enumerate(self['faces']):
                entry = []
                for _f in f:
                    if 'texcoord_index' not in _f:
                        if i > 0:
                            warnings.warn(("'texcoord_index' missing from face"
                                           + "%d, texcoord indices will be "
                                           + "ignored.") % i)
                        obj_ftexcoords = []
                        entry = []
                        break
                    entry.append(int(_f['texcoord_index']))
                if not entry:
                    break
                obj_ftexcoords.append(index_class(*entry))
            if obj_ftexcoords:
                kwargs['texCoordIndexList'] = array_class(obj_ftexcoords)
        # Normals
        if self.get('normals', []):
            obj_normals = []
            for n in self['normals']:
                obj_normals.append(pgl.Vector3(np.float64(n['i']),
                                               np.float64(n['j']),
                                               np.float64(n['k'])))
            kwargs['normalList'] = pgl.Point3Array(obj_normals)
            obj_fnormals = []
            for i, f in enumerate(self['faces']):
                entry = []
                for _f in f:
                    if 'normal_index' not in _f:
                        if i > 0:
                            warnings.warn(("'normal_index' missing from face"
                                           + "%d, normal indices will be "
                                           + "ignored.") % i)
                        obj_fnormals = []
                        entry = []
                        break
                    entry.append(int(_f['normal_index']))
                if not entry:
                    break
                obj_fnormals.append(index_class(*entry))
            if obj_fnormals:
                kwargs['normalIndexList'] = array_class(obj_fnormals)
        return smb_class, args, kwargs

    def append(self, solf):
        r"""Append new ply information to this dictionary.

        Args:
            solf (ObjDict): Another ply to append to this one.

        """
        exist_map = {'vertex_index': len(self.get('vertices', [])),
                     'texcoord_index': len(self.get('texcoords', [])),
                     'normal_index': len(self.get('normals', [])),
                     'param_index': len(self.get('params', []))}
        exist_map.update(points=exist_map['vertex_index'],
                         curve2Ds=exist_map['param_index'])
        # Vertex fields
        for k in ['vertices', 'texcoords', 'normals', 'params']:
            if k in solf:
                if k not in self:
                    self[k] = []
                self[k] += solf[k]
        # Points/2D curves
        for k in ['points', 'curve2Ds']:
            if k in solf:
                if k not in self:
                    self[k] = []
                for x in solf[k]:
                    self[k].append([v + exist_map[k] for v in x])
        # Face/line fields
        for k in ['lines', 'faces']:
            if k in solf:
                if k not in self:
                    self[k] = []
                for x in solf[k]:
                    iele = [{ik: v[ik] + exist_map[ik] for ik in v.keys()} for v in x]
                    self[k].append(iele)
        # Curves
        k = 'curves'
        if k in solf:
            if k not in self:
                self[k] = []
            for x in solf[k]:
                iele = copy.deepcopy(x)
                iele['vertex_indices'] = [v + exist_map['vertex_index']
                                          for v in x['vertex_indices']]
        # Surfaces
        k = 'surfaces'
        if k in solf:
            if k not in self:
                self[k] = []
            for x in solf[k]:
                iele = copy.deepcopy(x)
                iele['vertex_indices'] = [{ik: v[ik] + exist_map[ik] for ik in v.keys()}
                                          for v in x['vertex_indices']]
        # Merge material using first in list
        material = None
        for x in [self, solf]:
            if x.get('material', None) is not None:
                material = x['material']
                break
        if material is not None:
            self['material'] = material
        return self

    def apply_scalar_map(self, *args, **kwargs):
        r"""Set the color of faces in a 3D object based on a scalar map.
        This creates a copy unless no_copy is True.

        Args:
            scalar_arr (arr): Scalar values that should be mapped to colors
                for each face.
            color_map (str, optional): The name of the color map that should
                be used. Defaults to 'plasma'.
            vmin (float, optional): Value that should map to the minimum of the
                colormap. Defaults to min(scalar_arr).
            vmax (float, optional): Value that should map to the maximum of the
                colormap. Defaults to max(scalar_arr).
            scaling (str, optional): Scaling that should be used to map the scalar
                array onto the colormap. Defaults to 'linear'.
            scale_by_area (bool, optional): If True, the elements of the scalar
                array will be multiplied by the area of the corresponding face.
                If True, vmin and vmax should be in terms of the scaled array.
                Defaults to False.
            no_copy (bool, optional): If True, the returned object will not be a
                copy. Defaults to False.

        Returns:
            dict: Obj with updated vertex colors.

        """
        kwargs['_as_obj'] = True
        return super(ObjDict, self).apply_scalar_map(*args, **kwargs)
    

# The base class could be anything since it is discarded during registration,
# but is set to JSONObjectMetaschemaType here for transparancy since this is
# what the base class is determined to be on loading the schema
class ObjMetaschemaType(JSONObjectMetaschemaType):
    r"""Obj 3D structure map."""

    _empty_msg = {'vertices': [], 'faces': []}
    python_types = (dict, ObjDict)
    schema_file = _schema_file
    _replaces_existing = False

    @classmethod
    def _encode_object_property(cls, obj, order, req_keys=False):
        if req_keys:
            sep = '/'
        else:
            sep = ' '
        plist = []
        if isinstance(obj, (list, tuple)):
            for x in obj:
                plist.append(cls._encode_object_property(x, order, req_keys=True))
            return sep.join(plist)
        elif isinstance(obj, dict):
            for i, k in enumerate(order):
                if isinstance(k, dict):
                    assert(len(k) == 1)
                    ksub = list(k.keys())[0]
                    if ksub in obj:
                        plist.append(cls._encode_object_property(obj[ksub], k[ksub]))
                elif isinstance(k, (list, tuple)):
                    assert(len(k) == 1)
                    ksub = k[0]
                    if ksub in obj:
                        plist.append(cls._encode_object_property(obj[ksub], ksub))
                else:
                    if k in obj:
                        plist.append(cls._encode_object_property(obj[k], k))
                    elif req_keys:
                        plist.append('')
            return sep.join(plist)
        else:
            if order in _index_properties:
                # Add one at write to indexes as .obj is not zero indexed
                return _default_property_formats[order] % (obj + 1)
            else:
                return _default_property_formats[order] % obj

    @classmethod
    def _decode_object_property(cls, values, order):
        if isinstance(values, (list, tuple)):
            if not isinstance(order, (list, tuple)):
                out = [cls._decode_object_property(v, order) for v in values]
            elif (len(values) > 0) and ('/' in values[0]):
                out = [cls._decode_object_property(v, order) for v in values]
            else:
                out = {}
                for i, (o, v) in enumerate(zip(order, values)):
                    if not v:
                        continue
                    if isinstance(o, dict):
                        assert(len(o) == 1)
                        osub = list(o.keys())[0]
                        out[osub] = cls._decode_object_property(values[i:], o[osub])
                        break
                    elif isinstance(o, (list, tuple)):
                        assert(len(o) == 1)
                        osub = o[0]
                        out[osub] = cls._decode_object_property(values[i:], osub)
                    else:
                        out[o] = cls._decode_object_property(v, o)
        else:
            if not isinstance(order, (list, tuple)):
                ftranslate = _default_property_converters[order]
                out = ftranslate(values)
                if order in _index_properties:
                    # Subtract 1 from indexes because .obj is not zero indexed
                    out -= 1
            else:
                assert('/' in values)
                subvalues = values.split('/')
                assert(len(order) == len(subvalues))
                out = cls._decode_object_property(subvalues, order)
        return out

    @classmethod
    def encode_data(cls, obj, typedef, comments=[], newline='\n'):
        r"""Encode an object's data.

        Args:
            obj (object): Object to encode.
            typedef (dict): Type definition that should be used to encode the
                object.
            comments (list, optional): List of comments that should be included in
                the file header. Defaults to lines describing the automated origin
                of the file.
            newline (str, optional): String that should be used to delineated end
                of lines. Defaults to '\n'.

        Returns:
            bytes, str: Serialized message.

        """
        # Encode header
        header = ['# Author ygg_auto',
                  '# Generated by yggdrasil']
        header += ['# ' + c for c in comments]
        header += ['']
        # Encode body
        body = []
        for e in _default_element_order:
            if (e not in obj):
                continue
            if (e == 'material'):
                body.append('%s %s' % (_map_element2code[e], obj['material']))
                continue
            for ie in obj[e]:
                ivalue = cls._encode_object_property(ie, _default_property_order[e])
                iline = '%s %s' % (_map_element2code[e], ivalue)
                body.append(iline.strip())  # Ensure trailing spaces are removed
        return newline.join(header + body) + newline
        
    @classmethod
    def encode_data_readable(cls, obj, typedef):
        r"""Encode an object's data in a readable format.

        Args:
            obj (object): Object to encode.
            typedef (dict): Type definition that should be used to encode the
                object.

        Returns:
            string: Encoded object.

        """
        return cls.encode_data(obj, typedef)
    
    @classmethod
    def decode_data(cls, msg, typedef):
        r"""Decode an object.

        Args:
            msg (string): Encoded object to decode.
            typedef (dict): Type definition that should be used to decode the
                object.

        Returns:
            object: Decoded object.

        """
        lines = backwards.as_str(msg).splitlines()
        metadata = {'comments': []}
        out = {}
        # Parse
        for line_count, line in enumerate(lines):
            if line.startswith('#'):
                metadata['comments'].append(line)
                continue
            values = line.split()
            if not values:
                continue
            if values[0] not in _map_code2element:
                raise ValueError("Type code '%s' on line %d not understood"
                                 % (values[0], line_count))
            e = _map_code2element[values[0]]
            if e not in out:
                out[e] = []
            if e in ['material']:
                out[e] = values[1]
                continue
            else:
                out[e].append(
                    cls._decode_object_property(values[1:], _default_property_order[e]))
        # Return
        # out.update(**metadata)
        return ObjDict(out)

    @classmethod
    def coerce_type(cls, obj, typedef=None, **kwargs):
        r"""Coerce objects of specific types to match the data type.

        Args:
            obj (object): Object to be coerced.
            typedef (dict, optional): Type defintion that object should be
                coerced to. Defaults to None.
            **kwargs: Additional keyword arguments are metadata entries that may
                aid in coercing the type.

        Returns:
            object: Coerced object.

        """
        if not backwards.PY2:
            try:
                obj['material'] = backwards.as_str(obj['material'])
            except BaseException:
                pass
        return super(ObjMetaschemaType, cls).coerce_type(obj, typedef=typedef,
                                                         **kwargs)

    @classmethod
    def updated_fixed_properties(cls, obj):
        r"""Get a version of the fixed properties schema that includes information
        from the object.

        Args:
            obj (object): Object to use to put constraints on the fixed properties
                schema.

        Returns:
            dict: Fixed properties schema with object dependent constraints.

        """
        out = super(ObjMetaschemaType, cls).updated_fixed_properties(obj)
        # Constrain dependencies for indexes into other elements
        depend_map = {'vertex_index': 'vertices', 'vertex_indices': 'vertices',
                      'texcoord_index': 'texcoords',
                      'normal_index': 'normals'}
        check_depends = {'lines': ['texcoord_index'],
                         'faces': ['texcoord_index', 'normal_index'],
                         'surfaces:vertex_indices': ['texcoord_index', 'normal_index']}
        for e, props in check_depends.items():
            sube = None
            if ':' in e:
                e, sube = e.split(':')
            if not ((e in obj) and isinstance(obj[e], (list, tuple))):
                continue
            req_flags = {k: False for k in props}
            for o in obj[e]:
                if sum(req_flags.values()) == len(props):
                    break
                if isinstance(o, dict):
                    assert(sube)
                    if (((sube not in o) or (not isinstance(o[sube], (list, tuple)))
                         or (len(o[sube]) == 0) or (not isinstance(o[sube][0], dict)))):
                        continue
                    for p in props:
                        if p in o[sube][0]:
                            req_flags[p] = True
                elif isinstance(o, (list, tuple)):
                    if (len(o) == 0) or (not isinstance(o[0], dict)):
                        continue
                    for p in props:
                        if p in o[0]:
                            req_flags[p] = True
            # Set dependencies
            for p in req_flags.keys():
                if not req_flags[p]:
                    continue
                if depend_map[p] not in out['dependencies'][e]:
                    out['dependencies'][e].append(depend_map[p])
        # Contrain indices on number of elements refered to
        if ('vertices' in obj) and isinstance(obj['vertices'], (list, tuple)):
            out['definitions']['curve']['properties']['vertex_indices']['items'][
                'maximum'] = len(obj['vertices']) - 1
        if ('params' in obj) and isinstance(obj['params'], (list, tuple)):
            out['definitions']['curve2D']['items']['maximum'] = len(obj['params']) - 1
        for e in ['line', 'face', 'surface']:
            if e == 'surface':
                iprop = out['definitions'][e]['properties']['vertex_indices'][
                    'items']['properties']
            else:
                iprop = out['definitions'][e]['items']['properties']
            for k, e_depends in depend_map.items():
                if k in iprop:
                    if (e_depends in obj) and isinstance(obj[e_depends], (list, tuple)):
                        iprop[k]['maximum'] = len(obj[e_depends]) - 1
        return out


ObjDict._type_class = ObjMetaschemaType



from yggdrasil.interface import YggInput, YggOutput


def calculate_growth(photosynthesis_rate):
    r"""Calculate the plant growth rate from the photosynthesis rate.

    Args:
        photosynthesis_rate (float): Rate of photosynthesis.

    Returns:
        float: Growth rate.

    """
    return 0.5 * photosynthesis_rate


if __name__ == '__main__':
    input = YggInput('photosynthesis_rate')
    output = YggOutput('growth_rate', '%f\n')
    
    while True:
        flag, prate = input.recv()
        if not flag:
            print('growth: No more input.')
            break
        grate = calculate_growth(*prate)
        print('growth: photosynthesis rate = %f ---> growth rate = %f' % (
            prate[0], grate))
        flag = output.send(grate)
        if not flag:
            raise RuntimeError('growth: Error sending growth rate.')



from yggdrasil.drivers.ClientRequestDriver import ClientRequestDriver


class ClientDriver(ClientRequestDriver):
    r"""Alias for ClientRequestDriver."""

    _connection_type = 'client'
    _schema_subtype_description = ('Connection between a model acting as a '
                                   'client and a server request comm.')



import numpy as np
import unittest
from yggdrasil import backwards
from yggdrasil.serialize import DefaultSerialize
from yggdrasil.serialize.tests import test_SerializeBase as parent


class TestFunctionalSerialize(parent.TestSerializeBase):
    r"""Test class for FunctionalSerialize."""

    _cls = 'FunctionalSerialize'
    testing_option_kws = {'as_format': True}
    
    def __init__(self, *args, **kwargs):
        super(TestFunctionalSerialize, self).__init__(*args, **kwargs)
        self.func_serialize = self._func_serialize
        self.func_deserialize = self._func_deserialize

    def get_options(self):
        r"""Get testing options."""
        out = {'kwargs': {'func_serialize': self.func_serialize,
                          'func_deserialize': self.func_deserialize},
               'empty': b'',
               'objects': [['one', np.int32(1), 1.0],
                           ['two', np.int32(2), 1.0]],
               'extra_kwargs': {},
               'typedef': {'type': 'bytes'},
               'dtype': None,
               'is_user_defined': True}
        return out
        
    def _func_serialize(self, args):  # pragma: no cover
        r"""Method that serializes using repr."""
        return backwards.as_bytes(repr(args))

    def _func_deserialize(self, args):  # pragma: no cover
        r"""Method that deserializes using eval."""
        if len(args) == 0:
            return self.testing_options['empty']
        x = eval(backwards.as_str(args))
        return x

    def test_serialize_sinfo(self):
        r"""Test serialize/deserialize with serializer info."""
        self.assert_raises(RuntimeError, self.instance.serialize,
                           self.testing_options['objects'][0],
                           add_serializer_info=True)
    

class FakeSerializer(DefaultSerialize.DefaultSerialize):
    r"""Fake serializer that mocks user defined serialization/deserialization
    routines."""

    _dont_register = True

    def func_serialize(self, args):  # pragma: no cover
        r"""Method that serializes using repr."""
        return backwards.as_bytes(repr(args))

    def func_deserialize(self, args):  # pragma: no cover
        r"""Method that deserializes using eval."""
        if len(args) == 0:
            return []
        x = eval(backwards.as_str(args))
        return x


class TestFunctionalSerialize_class(TestFunctionalSerialize):
    r"""Test class for FunctionalSerialize class with classes."""

    def get_options(self):
        r"""Get testing options."""
        temp_seri = FakeSerializer()
        assert(issubclass(temp_seri.__class__, DefaultSerialize.DefaultSerialize))
        out = super(TestFunctionalSerialize_class, self).get_options()
        out['kwargs'] = {'func_serialize': temp_seri,
                         'func_deserialize': temp_seri,
                         'encoded_datatype': {'type': 'bytes'}}
        return out
        

class TestFunctionalSerialize_error(TestFunctionalSerialize):
    r"""Test class for FunctionalSerialize class with incorrect functions."""

    def _func_serialize(self, args):
        r"""Method that serializes using repr."""
        return args

    def test_serialize(self):
        r"""Test serialize with function that dosn't return correct type."""
        self.assert_raises(TypeError, self.instance.serialize, (1,))

    @unittest.skipIf(True, 'Error testing')
    def test_serialize_no_metadata(self):
        r"""Test serializing without metadata."""
        pass  # pragma: no cover
        
    @unittest.skipIf(True, 'Error testing')
    def test_serialize_header(self):
        r"""Disabled: Test serialize/deserialize with header."""
        pass  # pragma: no cover

    @unittest.skipIf(True, 'Error testing')
    def test_serialize_sinfo(self):
        r"""Disabled: Test serialize/deserialize with serializer info."""
        pass  # pragma: no cover

    @unittest.skipIf(True, 'Error testing')
    def test_field_specs(self):
        r"""Disabled: Test field specifiers."""
        pass  # pragma: no cover



from yggdrasil.metaschema.datatypes import encode_type, compare_schema
from yggdrasil.metaschema.properties.MetaschemaProperty import MetaschemaProperty


class TemptypeMetaschemaProperty(MetaschemaProperty):
    r"""Property class for 'temptype' property."""

    name = 'temptype'
    schema = {'description': 'The type of the data for a single message.',
              'type': 'schema'}

    @classmethod
    def encode(cls, instance, typedef=None):
        r"""Encoder for the 'temptype' property."""
        return encode_type(instance, typedef)

    @classmethod
    def compare(cls, prop1, prop2, root1=None, root2=None):
        r"""Comparison for the 'temptype' property."""
        for e in compare_schema(prop1, prop2, root1=root1, root2=root2):
            yield e

    @classmethod
    def validate(cls, validator, value, instance, schema):
        r"""Validator for JSON schema validation of an instance by this property.
        If there is not a user provided validate function, the instance will be
        encoded and then the encoded value will be checked against the provided
        value using cls.compare.

        Args:
            validator (jsonschmea.Validator): JSON schema validator.
            value (object): Value of the property in the schema.
            instance (object): Instance to validate.
            schema (dict): Schema that instance should be validated against.
            
        Yields:
            str: Error messages associated with failed validation.

        """
        for error in validator.descend(instance, value):
            yield error



from yggdrasil.tests import YggTestClassInfo, assert_equal
from yggdrasil.metaschema import get_validator, get_metaschema
from yggdrasil.metaschema.datatypes import MetaschemaTypeError
from yggdrasil.metaschema.properties.MetaschemaProperty import (
    create_property)


def test_dynamic():
    r"""Test dynamic creation of property."""

    def encode(instance, typedef=None):
        return None

    def validate(validator, value, instance, schema):
        return

    def compare(prop1, prop2):
        if not prop2:
            yield 'Test error'
        return

    new_prop = create_property('invalid', None, encode, validate, compare,
                               dont_register=True)
    assert_equal(new_prop.encode('hello'), None)
    assert_equal(list(new_prop.validate(None, None, None, None)), [])
    assert_equal(list(new_prop.compare(True, True)), [])
    assert_equal(list(new_prop.compare(True, False)), ['Test error'])


class TestMetaschemaProperty(YggTestClassInfo):
    r"""Test class for MetaschemaProperty class."""
    
    _mod = 'MetaschemaProperty'
    _cls = 'MetaschemaProperty'
    
    def __init__(self, *args, **kwargs):
        super(TestMetaschemaProperty, self).__init__(*args, **kwargs)
        self._valid = []
        self._invalid = []
        self._encode_errors = []
        self._valid_compare = [(0, 0)]
        self._invalid_compare = [(0, 1)]
        self._valid_normalize_schema = []
        self.validator = get_validator()(get_metaschema())

    @property
    def mod(self):
        r"""str: Absolute name of module containing class to be tested."""
        return 'yggdrasil.metaschema.properties.%s' % self._mod

    @property
    def inst_args(self):
        r"""dict: Keyword arguments for creating a class instance."""
        return tuple()

    def test_encode(self):
        r"""Test encode method for the class."""
        for (instance, value) in self._valid:
            x = self.import_cls.encode(instance)
            errors = list(self.import_cls.compare(x, value))
            assert(not errors)
        if self.import_cls.name == 'base':
            self.assert_raises(NotImplementedError, self.import_cls.encode, None)

    def test_encode_errors(self):
        r"""Test errors raised by encode."""
        for instance in self._encode_errors:
            self.assert_raises(MetaschemaTypeError, self.import_cls.encode, instance)

    def test_validate_valid(self):
        r"""Test validation method for the class on valid objects."""
        validator = self.validator
        for (instance, value) in self._valid:
            schema = {self.import_cls.name: value}
            errors = list(
                self.import_cls.wrapped_validate(validator, value, instance, schema))
            assert(not errors)
        # Instances not of the associate type validate as true (skipped)
        if self.import_cls.name == 'base':
            errors = list(
                self.import_cls.wrapped_validate(validator, None, None, {}))
        assert(not errors)

    def test_validate_invalid(self):
        r"""Test validation method for the class on invalid objects."""
        validator = self.validator
        for (instance, value) in self._invalid:
            schema = {self.import_cls.name: value}
            errors = list(
                self.import_cls.wrapped_validate(validator, value, instance, schema))
            assert(errors)

    def test_compare_valid(self):
        r"""Test comparision method for the class on valid objects."""
        for x in self._valid_compare:
            errors = list(self.import_cls.compare(*x))
            assert(not errors)

    def test_compare_invalid(self):
        r"""Test comparision method for the class on invalid objects."""
        for x in self._invalid_compare:
            errors = list(self.import_cls.compare(*x))
            assert(errors)

    def test_normalize_in_schema(self):
        r"""Test normalization in schema."""
        for x, y in self._valid_normalize_schema:
            self.assert_equal(self.import_cls.normalize_in_schema(x), y)



import numpy as np
from yggdrasil.communication.tests import test_AsciiTableComm as parent


class TestPandasFileComm(parent.TestAsciiTableComm):
    r"""Test for PandasFileComm communication class."""

    comm = 'PandasFileComm'
    testing_option_kws = {'not_as_frames': True}


class TestPandasFileComm_nonames(TestPandasFileComm):
    r"""Test for PandasFileComm communication class without field names sent."""

    testing_option_kws = {'not_as_frames': True, 'no_names': True}


class TestPandasFileComm_single(TestPandasFileComm):
    r"""Test for PandasFileComm communication class with field names sent."""

    def get_options(self):
        r"""Get testing options."""
        nele = 5
        dtype = np.dtype(dict(formats=['float'], names=['f0']))
        arr = np.zeros((nele, ), dtype)
        out = {'kwargs': {},
               'contents': (b'f0\n' + 2 * nele * b'0.0\n'),
               'send': [[arr['f0']], [arr['f0']]],
               'recv': [[np.hstack([arr, arr])['f0']]],
               'recv_partial': [[[arr['f0']]], [[np.hstack([arr, arr])['f0']]]],
               'dict': {'f0': arr['f0']}}
        out['msg'] = out['send'][0]
        out['msg_array'] = arr
        return out

    def test_send_dict_default(self):
        r"""Test automated conversion of dictionary to pandas data frame."""
        self.do_send_recv(msg_send=self.testing_options['dict'],
                          msg_recv=self.testing_options['msg'])



import copy
from yggdrasil.communication.tests import test_CommBase


class TestAsyncComm(test_CommBase.TestCommBase):
    r"""Tests for AsyncComm communication class."""

    comm = 'AsyncComm'
    attr_list = (copy.deepcopy(test_CommBase.TestCommBase.attr_list)
                 + ['dont_backlog', 'backlog_send_ready',
                    'backlog_recv_ready'])

    def test_send_recv_after_close(self):
        r"""Test that send/recv after close returns false."""
        super(TestAsyncComm, self).test_send_recv_after_close()
        self.assert_equal(self.send_instance.n_msg_direct_send, 0)
        self.assert_equal(self.recv_instance.n_msg_direct_recv, 0)
        self.assert_equal(self.send_instance.n_msg_direct, 0)
        self.assert_equal(self.recv_instance.n_msg_direct, 0)

    def test_send_recv_direct(self):
        r"""Test send/recv direct."""
        self.send_instance.n_msg_backlog
        self.recv_instance.n_msg_backlog
        self.send_instance.stop_backlog()
        self.recv_instance.stop_backlog()
        self.do_send_recv(send_kwargs={'no_confirm': True},
                          recv_kwargs={'no_confirm': True})



import copy


_conversion_registry = {}


def register_conversion(from_type, to_type):
    r"""Register a conversion function for going between types.

    Args:
        from_type (str): Name of type being converted from.
        to_type (str): Name of type being converted to.

    Returns:
        function: Decorator that will register the conversion.

    """
    global _conversion_registry
    key = (from_type, to_type)
    if key in _conversion_registry:
        raise ValueError("Conversion '%s' to '%s' already in registry." % key)

    def decorator(function):
        _conversion_registry[key] = function
        return function

    return decorator


def get_conversion(from_type, to_type):
    r"""Get a conversion function for moving from one type to the other.

    Args:
        from_type (str): Name of type being converted from.
        to_type (str): Name of type being converted to.

    Returns:
         function: Method for performing the conversion. If there is no known
             conversion between the specified types, None will be returned.

    """
    key = (from_type, to_type)
    return _conversion_registry.get(key, None)


@register_conversion('ply', 'obj')
def ply2obj(ply):
    r"""Convert a Ply object to an Obj object.

    Args:
        ply (dict): Ply type object.

    Returns:
        dict: Obj data container.

    """
    out = {'material': ply['material']}
    # Vertices
    if 'vertices' in ply:
        out['vertices'] = copy.deepcopy(ply['vertices'])
    # Faces
    if 'faces' in ply:
        out['faces'] = []
        for f in ply['faces']:
            out['faces'].append([{'vertex_index': q} for q in f['vertex_index']])
    # Edges
    if 'edges' in ply:
        out['lines'] = []
        for e in ply['edges']:
            x = [{'vertex_index': e['vertex%d' % q]} for q in [1, 2]]
            out['lines'].append(x)
    return out


@register_conversion('obj', 'ply')
def obj2ply(obj):
    r"""Convert an Obj object to a Ply object.

    Args:
        obj (dict): Obj type object.

    Returns:
        dict: Ply data container.

    """
    out = {'material': obj['material']}
    # Vertices
    if 'vertices' in obj:
        out['vertices'] = copy.deepcopy(obj['vertices'])
        for v in out['vertices']:
            if 'w' in v:
                del v['w']
    # Faces
    if 'faces' in obj:
        out['faces'] = []
        for f in obj['faces']:
            out['faces'].append({'vertex_index': [x['vertex_index'] for x in f]})
    # Edges
    if 'lines' in obj:
        out['edges'] = []
        for e in obj['lines']:
            # Ply dosn't have arbitrary edges
            for ii in range(len(e) - 1):
                iedge = {'vertex1': e[ii]['vertex_index'],
                         'vertex2': e[ii + 1]['vertex_index']}
                out['edges'].append(iedge)
    return out



from yggdrasil import backwards
from yggdrasil.metaschema.encoder import encode_yaml, decode_yaml
from yggdrasil.serialize.JSONSerialize import JSONSerialize


class YAMLSerialize(JSONSerialize):
    r"""Class for serializing a python object into a bytes message using YAML.

    Args:
        indent (str, int, optional): String or number of spaces that should be
            used to indent each level within the seiralized structure. Defaults
            to '\t'.
        encoding (str, optional): Encoding that should be used to serialize the
            object. Defaults to 'utf-8'.
        default_flow_style (bool, optional): If True, nested collections will
            be serialized in the block style. If False, they will always be
            serialized in the flow style. See
            `PyYAML Documentation <https://pyyaml.org/wiki/PyYAMLDocumentation>`_.

    """

    _seritype = 'yaml'
    _schema_subtype_description = ('Serializes Python objects using the YAML '
                                   'standard.')
    _schema_properties = {
        'indent': {'type': ['string', 'int'], 'default': '\t'},
        'encoding': {'type': 'string', 'default': 'utf-8'},
        'default_flow_style': {'type': 'boolean', 'default': False}}
    _schema_excluded_from_inherit = ['sort_keys']

    def func_serialize(self, args):
        r"""Serialize a message.

        Args:
            args (obj): Python object to be serialized.

        Returns:
            bytes, str: Serialized message.

        """
        # Convert bytes to str because JSON cannot serialize bytes by default
        args = backwards.as_str(args, recurse=True, allow_pass=True)
        return encode_yaml(args, indent=self.indent, encoding=self.encoding,
                           default_flow_style=self.default_flow_style)

    def func_deserialize(self, msg):
        r"""Deserialize a message.

        Args:
            msg (str, bytes): Message to be deserialized.

        Returns:
            obj: Deserialized Python object.

        """
        return decode_yaml(msg)

    @classmethod
    def get_testing_options(cls, **kwargs):
        r"""Method to return a dictionary of testing options for this class.

        Returns:
            dict: Dictionary of variables to use for testing.

        """
        # iobj = {'a': ['b', int(1), float(1.0)], 'c': {'z': 'hello'}}
        iobj1 = {'a': ['b', int(1), float(1.0)], 'c': {'z': 'hello'}}
        iobj2 = {'d': 'new field'}
        # iobj3 = int(2)
        # iobj4 = [float(2.0)]
        out = {'kwargs': {},
               'empty': {}, 'dtype': None,
               'extra_kwargs': {},
               'objects': [iobj1, iobj2],  # , iobj3, iobj4],
               'typedef': {'type': 'object'}}
        out['contents'] = (b'a:\n- b\n- 1\n- 1.0\n'
                           b'c:\n    z: hello\n'
                           b'd: new field\n')
        # out['contents'] = (b'-   a:\n    - b\n    - 1\n    - 1.0\n'
        #                    b'    c:\n        z: hello\n'
        #                    b'    d: new field\n'
        #                    b'- 2\n- 2.0\n')
        return out



import os
import unittest
from yggdrasil import backwards
from yggdrasil.tests import assert_equal
from yggdrasil.communication import AsciiTableComm
from yggdrasil.communication.tests import test_AsciiFileComm as parent


def test_AsciiTableComm_nofmt():
    r"""Test read of asciitable without format."""
    test_file = os.path.join(os.getcwd(), 'temp_file.txt')
    rows = [('one', 1, 1.0), ('two', 2, 2.0), ('three', 3, 3.0)]
    lines = [backwards.format_bytes('%5s\t%d\t%f\n', r) for r in rows]
    contents = backwards.as_bytes(''.join(lines))
    with open(test_file, 'wb') as fd:
        fd.write(contents)
    inst = AsciiTableComm.AsciiTableComm('test', test_file, direction='recv')
    inst.open()
    for ans in rows:
        flag, x = inst.recv_dict()
        assert(flag)
        irow = [e for e in ans]
        irow[0] = backwards.as_bytes(irow[0])
        idict = {'f%d' % i: irow[i] for i in range(len(irow))}
        # irow = tuple(irow)
        assert_equal(x, idict)
    flag, x = inst.recv()
    assert(not flag)
    inst.close()
    os.remove(test_file)


class TestAsciiTableComm(parent.TestAsciiFileComm):
    r"""Test for AsciiTableComm communication class."""

    comm = 'AsciiTableComm'
    
    @unittest.skipIf(True, 'Table comm')
    def test_send_recv_comment(self):
        r"""Disabled: Test send/recv with commented message."""
        pass  # pragma: no cover


class TestAsciiTableComm_AsArray(TestAsciiTableComm):
    r"""Test for AsciiTableComm communication class."""

    testing_option_kws = {'array_columns': True}



import yggdrasil.drivers.tests.test_InterpretedModelDriver as parent


class TestPythonModelParam(parent.TestInterpretedModelParam):
    r"""Test parameters for PythonModelDriver."""

    driver = "PythonModelDriver"


class TestPythonModelDriverNoInit(TestPythonModelParam,
                                  parent.TestInterpretedModelDriverNoInit):
    r"""Test runner for PythonModelDriver without init."""
    pass


class TestPythonModelDriverNoStart(TestPythonModelParam,
                                   parent.TestInterpretedModelDriverNoStart):
    r"""Test runner for PythonModelDriver without start."""
    pass


class TestPythonModelDriver(TestPythonModelParam,
                            parent.TestInterpretedModelDriver):
    r"""Test runner for PythonModelDriver."""
    pass



from yggdrasil.drivers.ServerRequestDriver import ServerRequestDriver


class ServerDriver(ServerRequestDriver):
    r"""Alias for ServerRequestDriver."""

    _connection_type = 'server'
    _schema_subtype_description = ('Connection between a server request comm '
                                   'and a model acting as a server.')



import numpy as np
from yggdrasil.metaschema.datatypes.tests import (
    test_ScalarMetaschemaType as parent)


class TestOneDArrayMetaschemaType(parent.TestScalarMetaschemaType):
    r"""Test class for ArrayMetaschemaType class."""
    _mod = 'ArrayMetaschemaType'
    _cls = 'OneDArrayMetaschemaType'
    _shape = 10

    def __init__(self, *args, **kwargs):
        super(TestOneDArrayMetaschemaType, self).__init__(*args, **kwargs)
        self._valid_encoded[0]['length'] = len(self._array)
        self._valid_decoded.append(np.array([], self._array.dtype))

    def assert_result_equal(self, x, y):
        r"""Assert that serialized/deserialized objects equal."""
        np.testing.assert_array_equal(x, y)
        

class TestNDArrayMetaschemaType(parent.TestScalarMetaschemaType):
    r"""Test class for ArrayMetaschemaType class with 2D array."""
    _mod = 'ArrayMetaschemaType'
    _cls = 'NDArrayMetaschemaType'
    _shape = (4, 5)

    def __init__(self, *args, **kwargs):
        super(TestNDArrayMetaschemaType, self).__init__(*args, **kwargs)
        self._valid_encoded[0]['shape'] = list(self._array.shape)

    def assert_result_equal(self, x, y):
        r"""Assert that serialized/deserialized objects equal."""
        np.testing.assert_array_equal(x, y)



import os
import sys
import subprocess
PY_MAJOR_VERSION = sys.version_info[0]


def install_matlab(as_user=False):
    r"""Attempt to install the MATLAB engine API for Python.

    Arguments:
        as_user (bool, optional): If True, the install will be called with
            --user for a local install. Defaults to False.

    Returns:
        bool: True if install succeded, False otherwise.

    """
    # Check to see if its already installed
    try:
        import matlab.engine
        return True
    except ImportError:
        pass
    # Get location of matlab root
    mtl_id = '=MATLABROOT='
    cmd = "fprintf('" + mtl_id + "%s" + mtl_id + "', matlabroot); exit();"
    mtl_cmd = ['matlab', '-nodisplay', '-nosplash', '-nodesktop', '-nojvm',
               '-r', '%s' % cmd]
    try:
        mtl_proc = subprocess.check_output(mtl_cmd)
        if PY_MAJOR_VERSION == 3:
            mtl_proc = mtl_proc.decode("utf-8")
    except BaseException:
        return False
    if mtl_id not in mtl_proc:
        return False
    mtl_root = mtl_proc.split(mtl_id)[-2]
    # Install engine API
    mtl_setup = os.path.join(mtl_root, 'extern', 'engines', 'python')
    if not os.path.isdir(mtl_setup):
        return False
    blddir = os.path.join(os.path.expanduser('~'), 'matlab_python_api')
    if not os.path.isdir(blddir):
        os.mkdir(blddir)
    cmd = [sys.executable, 'setup.py',
           'build', '--build-base=%s' % blddir,
           'install']
    if '--user' in sys.argv:
        cmd.append('--user')
    try:
        result = subprocess.check_output(cmd, cwd=mtl_setup)
        if PY_MAJOR_VERSION == 3:
            result = result.decode("utf-8")
        print(result)
    except subprocess.CalledProcessError:
        return False
    return True


if __name__ == "__main__":
    out = install_matlab(as_user=('--user' in sys.argv))
    if out:
        print("MATLAB engine installed.")
    else:
        raise Exception("Failed to install MATLAB engine.")



import os
import sys
PY_MAJOR_VERSION = sys.version_info[0]
IS_WINDOWS = (sys.platform in ['win32', 'cygwin'])
# Import config parser
try:
    from ConfigParser import RawConfigParser as HandyConfigParser
except ImportError:
    try:
        from configparser import RawConfigParser as HandyConfigParser
    except ImportError:
        HandyConfigParser = None


def add_excl_rule(excl_list, new_rule):
    r"""Add an exclusion rule to the list if its not in there.

    Args:
        excl_list (list): List of exclusion rules to update.
        new_rule (str): New rule to add to the list if it already exists.

    Returns:
        list: Updated exclustion rules.

    """
    if new_rule not in excl_list:
        excl_list.append(new_rule)
    return excl_list


def rm_excl_rule(excl_list, new_rule):
    r"""Remove an exclusion rule from the list if its in there.

    Args:
        excl_list (list): List of exclusion rules to update.
        new_rule (str): New rule to remove from the list if it exists.

    Returns:
        list: Updated exclustion rules.

    """
    if new_rule in excl_list:
        excl_list.remove(new_rule)
    return excl_list


def create_coveragerc(matlab_installed=False, lpy_installed=False):
    r"""Create the coveragerc to reflect the OS, Python version, and availability
    of matlab. Parameters from the setup.cfg file will be added. If the
    .coveragerc file already exists, it will be read first before adding setup.cfg
    options.
    

    Args:
        matlab_installed (bool, optional): Truth of if matlab is installed or not.
            Defaults to False.
        lpy_installed (bool, optional): Truth of if lpy is installed or not.
            Defaults to False.

    Returns:
        bool: True if the file was created/updated successfully, False otherwise.

    """
    if HandyConfigParser is None:
        return False
    covrc = os.path.join(os.path.dirname(__file__), '.coveragerc')
    cp = HandyConfigParser("")
    # Read from existing .coveragerc
    if os.path.isfile(covrc):
        cp.read(covrc)
    # Read options from setup.cfg
    setup_cfg = os.path.join(os.path.dirname(__file__), 'setup.cfg')
    cp_cfg = HandyConfigParser("")
    cp_cfg.read(setup_cfg)
    # Transfer options
    for x in cp_cfg.sections():
        if x.startswith('coverage:'):
            sect_cp = x.split('coverage:')[-1]
            if not cp.has_section(sect_cp):
                cp.add_section(sect_cp)
            for opt in cp_cfg.options(x):
                if cp.has_option(sect_cp, opt):
                    val_old = [l.strip() for l in cp.get(sect_cp, opt).split('\n')]
                    val_new = [l.strip() for l in cp_cfg.get(x, opt).split('\n')]
                    for v in val_new:
                        if v not in val_old:
                            val_old.append(v)
                    opt_new = '\n'.join(val_old)
                else:
                    opt_new = cp_cfg.get(x, opt)
                cp.set(sect_cp, opt, opt_new)
    # Exclude rules for all files
    if not cp.has_section('report'):
        cp.add_section('report')
    if cp.has_option('report', 'exclude_lines'):
        excl_str = cp.get('report', 'exclude_lines')
        excl_list = excl_str.strip().split('\n')
    else:
        excl_list = []
    # Platform
    if IS_WINDOWS:
        excl_list = rm_excl_rule(excl_list, 'pragma: windows')
    else:
        excl_list = add_excl_rule(excl_list, 'pragma: windows')
    # Python version
    verlist = [2, 3]
    for v in verlist:
        vincl = 'pragma: Python %d' % v
        if PY_MAJOR_VERSION == v:
            excl_list = rm_excl_rule(excl_list, vincl)
        else:
            excl_list = add_excl_rule(excl_list, vincl)
    # Matlab
    if matlab_installed:
        excl_list = add_excl_rule(excl_list, 'pragma: no matlab')
        excl_list = rm_excl_rule(excl_list, 'pragma: matlab')
    else:
        excl_list = add_excl_rule(excl_list, 'pragma: matlab')
        excl_list = rm_excl_rule(excl_list, 'pragma: no matlab')
    # LPy
    if lpy_installed:
        excl_list = add_excl_rule(excl_list, 'pragma: no lpy')
        excl_list = rm_excl_rule(excl_list, 'pragma: lpy')
    else:
        excl_list = add_excl_rule(excl_list, 'pragma: lpy')
        excl_list = rm_excl_rule(excl_list, 'pragma: no lpy')
    # Add new rules
    cp.set('report', 'exclude_lines', '\n' + '\n'.join(excl_list))
    # Write
    with open(covrc, 'w') as fd:
        cp.write(fd)
    return True


if __name__ == "__main__":
    try:
        import matlab.engine
        matlab_installed = True
    except ImportError:
        matlab_installed = False
    try:
        from openalea import lpy
        lpy_installed = True
    except ImportError:
        lpy_installed = False
    flag = create_coveragerc(matlab_installed=matlab_installed,
                             lpy_installed=lpy_installed)
    if not flag:
        raise Exception("Failed to create/update converagerc file.")



from yggdrasil.tests import assert_raises, assert_equal
from yggdrasil.metaschema import datatypes
from yggdrasil.metaschema.tests import _valid_objects
from yggdrasil.metaschema.datatypes.ScalarMetaschemaType import (
    ScalarMetaschemaType)


def test_get_type_class():
    r"""Test get_type_class."""
    for v in _valid_objects.keys():
        datatypes.get_type_class(v)
    assert_raises(ValueError, datatypes.get_type_class, 'invalid')


def test_register_type_errors():
    r"""Test errors in register_type for duplicate."""
    assert_raises(ValueError, datatypes.register_type, ScalarMetaschemaType)
    type_args = ('FakeType', (ScalarMetaschemaType, ),
                 {'name': 'new', 'properties': ['invalid']})
    assert_raises(ValueError, type, *type_args)
    

def test_add_type_from_schema_errors():
    r"""Test errors in add_type_from_schema."""
    assert_raises(ValueError, datatypes.add_type_from_schema, 'invalid_file')


def test_get_type_from_def():
    r"""Test get_type_from_def."""
    datatypes.get_type_from_def('float')
    datatypes.get_type_from_def({'type': 'float'})
    datatypes.get_type_from_def({'a': 'float', 'b': 'int'})
    datatypes.get_type_from_def(['float', 'int'])
    assert_raises(TypeError, datatypes.get_type_from_def, None)


def test_guess_type_from_msg():
    r"""Test guess_type_from_msg."""
    assert_raises(ValueError, datatypes.guess_type_from_msg,
                  b'fake message')


def test_guess_type_from_obj():
    r"""Test guess_type_from_obj."""
    invalid_objects = [object()]  # , object()]
    for t, x in _valid_objects.items():
        assert_equal(datatypes.guess_type_from_obj(x).name, t)
    for x in invalid_objects:
        assert_raises(datatypes.MetaschemaTypeError,
                      datatypes.guess_type_from_obj, x)


def test_encode_decode():
    r"""Test encode/decode for valid objects."""
    for x in _valid_objects.values():
        y = datatypes.encode(x)
        z = datatypes.decode(y)
        assert_equal(z, x)
        t = datatypes.encode_type(x)
        d = datatypes.encode_data(x)
        w = datatypes.decode_data(d, t)
        assert_equal(w, x)
    assert_raises(ValueError, datatypes.decode_data, b'', None)


def test_encode_decode_readable():
    r"""Test encode_data_reable/decode for valid objects."""
    for x in _valid_objects.values():
        datatypes.encode_data_readable(x)
        

def test_compare_schema():
    r"""Test for compare_schema."""
    valid = [
        ({'type': 'int'}, {'type': 'int'}),
        ({'type': 'int'}, {'type': 'scalar', 'subtype': 'int'}),
        ({'type': 'scalar', 'subtype': 'int'}, {'type': 'int'}),
        ({'type': 'int', 'unnamed': 0}, {'type': 'int', 'unnamed': 1}),
        ({'type': 'int', 'unnamed': 0}, {'type': 'int'}),
        ({'type': 'object', 'definitions': {'a': {'type': 'int'}},
          'properties': {'x': {'$ref': '#/definitions/a'}}},
         {'type': 'object', 'definitions': {'b': {'type': 'int'}},
          'properties': {'x': {'$ref': '#/definitions/b'}}}),
        ({'type': 'object', 'properties': {'x': {'type': 'float'}}},
         {'type': 'object', 'properties': {'x': {'type': 'float'},
                                           'y': {'type': 'float'}},
          'required': ['x']})]
    invalid = [
        ({'type': 'int'}, {}), ({}, {'type': 'int'}),
        ({'type': 'int'}, {'type': 'int', 'precision': 4}),
        ({'type': 'object', 'definitions': {'a': {'type': 'float'}},
          'properties': {'x': {'$ref': '#/definitions/a'}}},
         {'type': 'object', 'definitions': {'b': {'type': 'int'}},
          'properties': {'x': {'$ref': '#/definitions/b'}}})]
    for x in valid:
        errors = list(datatypes.compare_schema(*x))
        assert(not errors)
    for x in invalid:
        errors = list(datatypes.compare_schema(*x))
        assert(errors)



import os
import unittest
from yggdrasil.tests import scripts, assert_raises, assert_equal
import yggdrasil.drivers.tests.test_CompiledModelDriver as parent
from yggdrasil.drivers.MakeModelDriver import MakeModelDriver, MakeCompiler


_driver_installed = MakeModelDriver.is_installed()


def test_MakeCompiler():
    r"""Test MakeCompiler class."""
    assert_equal(MakeCompiler.get_output_file(None, target='clean'), 'clean')


@unittest.skipIf(_driver_installed, "C Library installed")
def test_MakeModelDriver_no_C_library():  # pragma: windows
    r"""Test MakeModelDriver error when C library not installed."""
    assert_raises(RuntimeError, MakeModelDriver, 'test', scripts['make'])


@unittest.skipIf(not _driver_installed, "C Library not installed")
def test_MakeModelDriver_error_notarget():
    r"""Test MakeModelDriver error for invalid target."""
    makedir, target = os.path.split(scripts['make'])
    assert_raises(RuntimeError, MakeModelDriver, 'test', 'invalid',
                  makedir=makedir)


@unittest.skipIf(not _driver_installed, "C Library not installed")
def test_MakeModelDriver_error_nofile():
    r"""Test MakeModelDriver error for missing Makefile."""
    makedir, target = os.path.split(scripts['make'])
    assert_raises(RuntimeError, MakeModelDriver, 'test', 'invalid')


class TestMakeModelParam(parent.TestCompiledModelParam):
    r"""Test parameters for MakeModelDriver."""

    driver = 'MakeModelDriver'
    
    def __init__(self, *args, **kwargs):
        super(TestMakeModelParam, self).__init__(*args, **kwargs)
        self.attr_list += ['target', 'makedir', 'makefile']
        self.makedir, self.target = os.path.split(self.src[0])
        self.makefile = os.path.join(self.makedir, 'Makefile')
        self.args = [self.target]
        self._inst_kwargs['makefile'] = self.makefile
        if 'source_files' in self._inst_kwargs:
            del self._inst_kwargs['source_files']
        

class TestMakeModelDriverNoInit(TestMakeModelParam,
                                parent.TestCompiledModelDriverNoInit):
    r"""Test runner for MakeModelDriver without init."""
    pass


class TestMakeModelDriverNoStart(TestMakeModelParam,
                                 parent.TestCompiledModelDriverNoStart):
    r"""Test runner for MakeModelDriver without start."""
    
    def __init__(self, *args, **kwargs):
        super(TestMakeModelDriverNoStart, self).__init__(*args, **kwargs)
        # Version specifying makedir via working_dir
        self._inst_kwargs['yml']['working_dir'] = self.makedir
        del self._inst_kwargs['makefile']

    def test_compile_model(self):
        r"""Test compile model with alternate set of input arguments."""
        src = [self.target + '.c']
        self.instance.compile_model(target=self.target)
        self.instance.compile_model(source_files=src)
        self.assert_raises(RuntimeError, self.instance.compile_model,
                           source_files=src, target=self.target + 'invalid')
        

class TestMakeModelDriver(TestMakeModelParam, parent.TestCompiledModelDriver):
    r"""Test runner for MakeModelDriver."""
    pass



from yggdrasil import tools
from yggdrasil.config import ygg_cfg, locate_file
from yggdrasil.drivers.ModelDriver import ModelDriver


class InterpretedModelDriver(ModelDriver):
    r"""Base class for models written in interpreted languages.

    Args:
        name (str): Driver name.
        args (str or list): Argument(s) for running the model on the command
            line. This driver will check to see if there is an interpreter
            included in the args provided. If not, one will be added unless
            skip_interpreter is True.
        interpreter (str, optional): Name or path of interpreter executable that
            should be used to run the model. If not provided, the interpreter
            will be determined based on configuration options for the language
            (if present) and the default_interpreter class attribute.
        interpreter_flags (list, optional): Flags that should be passed to the
            interpreter when running the model. If not provided, the flags are
            determined based on configuration options for the language (if
            present) and the default_interpreter_flags class attribute.
        skip_interpreter (bool, optional): If True, no interpreter will be
            added to the arguments. This should only be used for subclasses
            that will not be invoking the model via the command line.
            Defaults to False.
        **kwargs: Additional keyword arguments are passed to the parent class.

    Class Attributes:
        default_interpreter (str): Name of interpreter that will be used if not
            set explicitly by instance or config file. Defaults to the language
            name if not set.
        default_interpreter_flags (list): Flags that will be passed to the
            interpreter when running the model by default if not set explicitly
            by instance or config file.

    Attributes:
        interpreter (str): Name or path to the interpreter that will be used.
        interpreter_flags (list): Flags that will be passed to the interpreter
            when running a model.

    """

    _schema_properties = {
        'interpreter': {'type': 'string'},
        'interpreter_flags': {'type': 'array', 'items': {'type': 'string'},
                              'default': []},
        'skip_interpreter': {'type': 'boolean', 'default': False}}
    executable_type = 'interpreter'
    default_interpreter = None
    default_interpreter_flags = []

    def __init__(self, name, args, **kwargs):
        super(InterpretedModelDriver, self).__init__(name, args, **kwargs)
        # Set defaults from attributes
        for k0 in ['interpreter']:
            for k in [k0, '%s_flags' % k0]:
                v = getattr(self, k, None)
                if v is None:
                    setattr(self, k, getattr(self, 'default_%s' % k))

    @staticmethod
    def after_registration(cls):
        r"""Operations that should be performed to modify class attributes after
        registration. For compiled languages this includes selecting the
        default compiler. The order of precedence is the config file 'compiler'
        option for the language, followed by the environment variable set by
        _compiler_env, followed by the existing class attribute.
        """
        ModelDriver.after_registration(cls)
        if cls.language is not None:
            for k in ['interpreter']:
                # Set attribute defaults based on config options
                for k0 in [k, '%s_flags' % k]:
                    ka = 'default_%s' % k0
                    if k0.endswith('_flags'):
                        old_val = getattr(cls, ka)
                        old_val += ygg_cfg.get(cls.language, k0, '').split()
                    else:
                        setattr(cls, ka, ygg_cfg.get(cls.language, k0,
                                                     getattr(cls, ka)))
            # Set default interpreter based on language
            if cls.default_interpreter is None:
                cls.default_interpreter = cls.language
                    
    def parse_arguments(self, *args, **kwargs):
        r"""Sort model arguments to determine which one is the executable
        and which ones are arguments.

        Args:
            *args: Arguments are passed to the parent class's method.
            **kwargs: Keyword arguments are passed to the parent class's method.

        """
        super(InterpretedModelDriver, self).parse_arguments(*args, **kwargs)
        self.model_src = self.model_file
        
    @classmethod
    def get_interpreter(cls):
        r"""Command required to run a model written in this language from
        the command line.

        Returns:
            str: Name of (or path to) interpreter executable.

        """
        out = getattr(cls, 'interpreter', getattr(cls, 'default_interpreter'))
        # out = None
        # if cls.language is not None:
        #     out = ygg_cfg.get(cls.language, 'interpreter',
        #                       getattr(cls, '_interpreter', cls.language))
        if out is None:
            raise NotImplementedError("Interpreter not set for language '%s'."
                                      % cls.language)
        return out

    @classmethod
    def get_interpreter_flags(cls):
        r"""Get the flags that should be passed to the interpreter when using it
        to run a model on the command line.

        Returns:
            list: Flags that should be passed to the interpreter on the command
                line.

        """
        out = getattr(cls, 'interpreter_flags',
                      getattr(cls, 'default_interpreter_flags'))
        return out

    @classmethod
    def language_executable(cls):
        r"""Command required to compile/run a model written in this language
        from the command line.

        Returns:
            str: Name of (or path to) compiler/interpreter executable required
                to run the compiler/interpreter from the command line.

        """
        return cls.get_interpreter()

    @classmethod
    def executable_command(cls, args, exec_type='interpreter', unused_kwargs={},
                           **kwargs):
        r"""Compose a command for running a program in this language with the
        provied arguments. If not already present, the interpreter command and
        interpreter flags are prepended to the provided arguments.

        Args:
            args (list): The program that returned command should run and any
                arguments that should be provided to it.
            exec_type (str, optional): Type of executable command that will be
                returned. If 'interpreter', a command using the interpreter is
                returned and if 'direct', the raw args being provided are
                returned. Defaults to 'interpreter'.
            unused_kwargs (dict, optional): Existing dictionary that unused
                keyword arguments should be added to. Defaults to {}.
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            list: Arguments composing the command required to run the program
                from the command line using the interpreter for this language.

        Raises:
            ValueError: If exec_type is not 'interpreter' or 'direct'.

        """
        ext = cls.language_ext
        assert(isinstance(ext, (tuple, list)))
        if exec_type == 'interpreter':
            # if (((cls.language not in args[0])
            if (((tools.which(args[0]) is None)
                 or any([args[0].endswith(e) for e in ext]))):
                args = [cls.get_interpreter()] + cls.get_interpreter_flags() + args
        elif exec_type != 'direct':
            raise ValueError("Invalid exec_type '%s'" % exec_type)
        unused_kwargs.update(kwargs)
        return args

    @classmethod
    def configure(cls, cfg):
        r"""Add configuration options for this language. This includes locating
        any required external libraries and setting option defaults.

        Args:
            cfg (YggConfigParser): Config class that options should be set for.

        Returns:
            list: Section, option, description tuples for options that could not
                be set.

        """
        out = super(InterpretedModelDriver, cls).configure(cfg)
        # Locate executable
        if not cls.is_language_installed():  # pragma: debug
            fpath = locate_file(cls.language_executable())
            if fpath:
                cfg.set(cls.language, 'interpreter', fpath)
        return out
